// Data_Reducer.java

import java.awt.*;
import java.awt.event.*;
import java.util.Timer;
import java.util.TimerTask;
import java.util.HashMap;
import java.util.Properties;
import java.util.Enumeration;
import java.util.Arrays.*;
import java.util.Date.*;
import java.util.TimeZone.*;
import java.util.regex.*;
import java.math.BigDecimal;


import javax.swing.*;
import javax.swing.event.*;
import javax.swing.text.*;
import ij.*;
import ij.io.*;
import java.nio.channels.FileChannel;
import java.io.RandomAccessFile;
import java.nio.channels.FileLock;
import java.nio.channels.OverlappingFileLockException;
import java.text.*;

import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeEvent;

import ij.process.*;
import ij.gui.*;
import ij.plugin.*;
import ij.plugin.frame.*;
import ij.plugin.filter.RankFilters.*;

import astroj.*;
import astroj.SpringUtil;		// OR GET IT FROM AN EXTERNAL PACKAGE
import astroj.AstroStackWindow.*;

import java.io.BufferedInputStream;
import java.io.DataInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;


/**
 * This plugin processes CCD image data, including bias and dark subtraction,
 * flat-field correction, and CCD-non-linearity correction. The plugin can process
 * existing datasets, or it can reduce data as it is being added to a directory 
 * (like directory watcher). This plugin is modeled after the directory_watcher
 * plugin written originally by F.V. Hessman, Inst. f. Astrophysik,
 * Georg-August-Unversitaet Goettingen, and later enhanced by K.A. Collins,
 * University of Louisville.
 * @author K.A. Collins, University of Louisville
 * @date 2010-11-16
 * @version 1.0
 * 
 */
public class Data_Processor implements PlugIn, ActionListener, ChangeListener, ItemListener, MouseListener, MouseMotionListener, MouseWheelListener, PropertyChangeListener
	{
	int pollingInterval = 0;		// IN SECONDS
	TimerTask	task = null;
	Timer		timer = null;
	boolean blocked = false;

	boolean onlyNew = false; // VERSION 1.1
    boolean firstRun = true;
    boolean restart = true;
    boolean loadNewCals = true;
    boolean useBeep = false;
    boolean requestStop = false;
    boolean sortNumerically = true;
    boolean updateExclude = true;
    boolean reloadFilenames = false;
    boolean usingNewWindow = false;
    boolean removing = false;

	int ignoredImages=0;

	String path = IJ.getDirectory ("home");
    String filepath = IJ.getDirectory ("home");
	String filenamePattern = "filenamepattern.fits";
    String oldFilenamePattern = "";
	String slash = "/";
    String biasPath = "", darkPath = "", flatPath = "";
    String biasFilePath = "", darkFilePath = "", flatFilePath = "";
    String mbiasPath = "", mdarkPath = "", mflatPath = "", sciencePath = "";
    String savePath = "", saveDirPath = "", saveFileName = "", saveFormat="";
    String numPrefix = "";
//    String raText = "", decText = "", latText = "", lonText = "";
    String s;

	File scienceDir, biasDirectory, darkDirectory, flatDirectory, biasMasterFile, darkMasterFile, flatMasterFile;
    File saveFile;
    ImagePlus scienceImp, mbiasImp, mdarkImp, mflatImp;
    String[] darkHeader = null, flatHeader = null, scienceHeader = null;
    double darkExpTime = 0, flatExpTime = 0, scienceExpTime = 0;
    int validFiles = 0, validBiasFiles = 0, validDarkFiles = 0, validFlatFiles = 0;
    int validMasterBiasFiles = 0, validMasterDarkFiles = 0, validMasterFlatFiles = 0;

	HashMap<String, String> images;
    ImagePlus openImage, testImp;
    int scienceHeight=0, scienceWidth=0, length = 0;
	int foundImages=0;
    int dialogFrameLocationX = 20;
    int dialogFrameLocationY = 20;
    long minFileNumber = 0, maxFileNumber = 1000000000;

    Class<?> imageWindowClass;
	protected boolean running = false;
	protected boolean ignoreAction = false;
    long flen, newflen;

	JFrame dialogFrame;
    Frame openFrame;
	JTextField dirText, filenamePatternText;
	JLabel statusLabel, pollingIntervalLabel, minFileNumberLabel, maxFileNumberLabel, validFilesLabel;
    JLabel validBiasFilesLabel, validDarkFilesLabel, validFlatFilesLabel, numPrefixLabel;
    JLabel validMasterBiasFilesLabel, validMasterDarkFilesLabel, validMasterFlatFilesLabel;
	JButton dirButton, fileButton, startButton, stopButton, clearButton;

	JCheckBox onlyNewBox;			
    JCheckBox useBeepBox, sortNumericallyBox;
    SpinnerModel pollingIntervalModel, minFileNumberModel, maxFileNumberModel;
    JSpinner pollingIntervalSpinner, minFileNumberSpinner, maxFileNumberSpinner;
    JCheckBox runPreMacroBox, runPostMacroBox, showLogBox, useGradientRemovalBox, useCosmicRemovalBox;
    JCheckBox preMacro1AutoLevelBox, postMacro1AutoLevelBox, postMacro2AutoLevelBox;
	JTextField preMacroText, postMacroText;
    JTextField raTextField, decTextField, latTextField, lonTextField;
	JButton preMacroButton, postMacroButton;

    boolean useBias, useDark, useFlat, useNLC, calcAirmass, saveAirmass;
    boolean createBias, createDark, createFlat, saveProcessedData;
    boolean biasMedian, darkMedian, flatMedian, saveFloatingPoint, scaleExpTime;
    boolean showMasters = false, showScience = true, showLog = true;
    boolean useGradientRemoval = false, useCosmicRemoval = false;
    String biasDir="", darkDir="", flatDir="", saveDir="";
    String biasBase="", darkBase="", flatBase="", saveSuffix="";
    String coeffCText="";
    double coeffA, coeffB, coeffC;
    double dec, ra, lat, lon;
    String biasMaster="", darkMaster="", flatMaster="";

    SimpleDateFormat f = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss");



    JCheckBox useBiasBox, useDarkBox, useFlatBox, useNLCBox, scaleExpTimeBox, updateExcludeBox;
    JCheckBox createBiasBox, createDarkBox, createFlatBox, saveProcessedDataBox;
    JCheckBox showMastersBox, showScienceBox, calcAirmassBox, saveAirmassBox;
    JRadioButton biasAverageRadio, darkAverageRadio, flatAverageRadio, saveIntegerRadio;
    JRadioButton biasMedianRadio, darkMedianRadio, flatMedianRadio, saveFloatingPointRadio;
    ButtonGroup biasRadioGroup, darkRadioGroup, flatRadioGroup, saveRadioGroup;
    JTextField biasDirField, darkDirField, flatDirField, saveDirField, numPrefixField;
    JTextField biasBaseField, darkBaseField, flatBaseField, saveSuffixField, saveFormatField;
    JSpinner coeffASpinner, coeffBSpinner, coeffCSpinner;
    SpinnerModel coeffAModel, coeffBModel, coeffCModel;
    JTextField biasMasterField, darkMasterField, flatMasterField;


	boolean runPreMacro = false;		
    boolean runPostMacro = false;
    boolean preMacro1AutoLevel = false;
    boolean postMacro1AutoLevel = false;
    boolean postMacro2AutoLevel = false;

	Macro_Runner runner = new Macro_Runner();
	String preMacroPath = new String(path);
    String postMacroPath = new String(path);
    String lastImageName = null;

    DecimalFormat twoPlaces = new DecimalFormat("0.00");
    DecimalFormat fourPlaces = new DecimalFormat("0.0000");
    DecimalFormat sixPlaces = new DecimalFormat("0.000000");
    DecimalFormat uptoFourPlaces = new DecimalFormat("0.####");
    DecimalFormat upto20Places = new DecimalFormat("0.####################");

    
    Canvas imc, openImageCanvas;
    ImageWindow win;
    Properties props;
    Enumeration enProps;
    String key;
    FileInfo imf;

    public void mouseClicked(MouseEvent e) {
    int x = e.getX();
    int y = e.getY();
    //int offscreenX = canvas.offScreenX(x);
    //int offscreenY = canvas.offScreenY(y);
    //IJ.write("mousePressed: "+offscreenX+","+offscreenY);
    }

    public void mousePressed(MouseEvent e) {}

    public void mouseReleased(MouseEvent e) {}

    public void mouseEntered(MouseEvent e) {}

    public void mouseExited(MouseEvent e) {}


    public void mouseDragged(MouseEvent e) {
        IJ.log("mouse dragged: "+e.getX()+","+e.getY());
        }
    public void mouseMoved(MouseEvent e) {
        IJ.log("mouse moved: "+e.getX()+","+e.getY());
        }




	public void run(String arg)
		{
        getPrefs();
		if (IJ.isWindows()) slash="\\";
        
        if (IJ.isWindows())
                {
                try {UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());}
//                try {UIManager.setLookAndFeel("javax.swing.plaf.metal.MetalLookAndFeel");}
//                try {UIManager.setLookAndFeel("com.sun.java.swing.plaf.windows.WindowsLookAndFeel");}
//                try {UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());}
                catch (Exception e) { }
                }
        else
                {
                try {UIManager.setLookAndFeel("javax.swing.plaf.metal.MetalLookAndFeel");}
                catch (Exception e) { }
                }
//            try {UIManager.setLookAndFeel("com.sun.java.swing.plaf.windows.WindowsLookAndFeel");}
//            try {UIManager.setLookAndFeel("com.sun.java.swing.plaf.nimbus.NimbusLookAndFeel");}
//            try {UIManager.setLookAndFeel("com.sun.java.swing.plaf.motif.MotifLookAndFeel");}


//        try {UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());}
//        catch (Exception e) { }
        images = new HashMap<String, String>(20);
//		startDialog();
//		}
//
//	/**
//	 * Constructs Swing interface.
//	 */
//	protected void startDialog ()
//		{

		dialogFrame = new JFrame ("CCD Data Processor");
        dialogFrame.setDefaultCloseOperation(javax.swing.WindowConstants.DISPOSE_ON_CLOSE);
        dialogFrame.addWindowListener(new WindowAdapter(){
            @Override
            public void windowClosing(WindowEvent e){
            saveAndClose();}});
		JPanel mainPanel = new JPanel (new SpringLayout());
		mainPanel.setBorder (BorderFactory.createEmptyBorder(10,10,10,10));

        int subpanelWidth = 180;
        if (IJ.isLinux()) subpanelWidth = 230;

        int subpanelHeight = 25;
        
        int boxWidthLeft = 75;
        int boxWidthRight = 125;
        int boxHeight = 24;

        Font p12 = new Font("Dialog",Font.PLAIN,12);
        Font b12 = new Font("Dialog",Font.BOLD,12);
        Font b11 = new Font("Dialog",Font.BOLD,11);
        Font b14 = new Font("Dialog",Font.BOLD,14);

        if (IJ.isLinux())
            {
            p12 = new Font("Dialog",Font.PLAIN,14);
            b12 = new Font("Dialog",Font.BOLD,14);
            b11 = new Font("Dialog",Font.BOLD,13);
            b14 = new Font("Dialog",Font.BOLD,16);
            }

        f.setTimeZone(java.util.TimeZone.getTimeZone("UTC"));

        JLabel imageLabel = new JLabel ("Science Image Selection: ");
        imageLabel.setFont(p12);
        imageLabel.setHorizontalAlignment (JTextField.RIGHT);
		mainPanel.add (imageLabel);

        JLabel imageDirLabel = new JLabel ("Primary Directory: ");
        imageDirLabel.setHorizontalAlignment (JTextField.RIGHT);
		mainPanel.add (imageDirLabel);

		dirText = new JTextField (path);
        scienceDir = new File (path);
        dirText.getDocument().addDocumentListener(new thisDocumentListener());
		dirText.setPreferredSize (new Dimension(300,subpanelHeight));
		dirText.setHorizontalAlignment (JTextField.LEFT);
        mainPanel.add (dirText);
                
		dirButton = new JButton ("Browse");
		dirButton.addActionListener (this);
        dirButton.setPreferredSize(new Dimension(100, 25));
		mainPanel.add (dirButton);

        JLabel fileLabel = new JLabel ("");
        fileLabel.setHorizontalAlignment (JTextField.LEFT);
		mainPanel.add (fileLabel);

        JLabel fileNameLabel = new JLabel ("Filename Pattern: ");
        fileNameLabel.setHorizontalAlignment (JTextField.RIGHT);
		mainPanel.add (fileNameLabel);

		JPanel fileNamePanel = new JPanel(new SpringLayout());
		filenamePatternText = new JTextField (filenamePattern);
		filenamePatternText.setPreferredSize (new Dimension(140,subpanelHeight));
        filenamePatternText.setToolTipText("Enter a science image filename pattern using ? and * as wildcard characters");
		filenamePatternText.setHorizontalAlignment (JTextField.RIGHT);
        filenamePatternText.getDocument().addDocumentListener(new thisDocumentListener());
		fileNamePanel.add (filenamePatternText);

		sortNumericallyBox = new JCheckBox("Sort Numerically",sortNumerically);
		sortNumericallyBox.addItemListener (this);
		fileNamePanel.add (sortNumericallyBox);

		updateExcludeBox = new JCheckBox("Update Exclude",updateExclude);
		updateExcludeBox.addItemListener (this);
        updateExcludeBox.setToolTipText("<html>"+"If checked, filename pattern through the last occurance of \"_\""+"<br>"+
                                             "will be copied to filename number filter \"Exclude\"."+"</html>");
		fileNamePanel.add (updateExcludeBox);

        SpringUtil.makeCompactGrid (fileNamePanel, 1, fileNamePanel.getComponentCount(), 6,6,6,6);
		mainPanel.add (fileNamePanel);

		fileButton = new JButton ("Browse");
		fileButton.addActionListener (this);
        fileButton.setPreferredSize(new Dimension(100, 25));
		mainPanel.add (fileButton);

        JLabel minValLabel = new JLabel ("");
        mainPanel.add (minValLabel);

        JLabel fileNumbersLabel = new JLabel ("Filename Number Filter:");
        fileNumbersLabel.setHorizontalAlignment (JLabel.RIGHT);
        mainPanel.add (fileNumbersLabel);

        JPanel minmaxValuePanel = new JPanel(new SpringLayout());

        numPrefixLabel = new JLabel("Exclude:");
        numPrefixLabel.setHorizontalAlignment (JLabel.RIGHT);
        minmaxValuePanel.add(numPrefixLabel);

        numPrefixField = new JTextField (numPrefix);
		numPrefixField.setPreferredSize (new Dimension(100,subpanelHeight));
		numPrefixField.setHorizontalAlignment (JTextField.LEFT);
        numPrefixField.setToolTipText("<html>"+"Defines the excluded filename prefix for filtering filenames numerically."+"<br>"+
                                             "Only numbers in the filename after the last occurance of this string"+"<br>"+
                                             "are used for filtering. Leave blank to use all numbers in the filename." +"</html>");
        numPrefixField.getDocument().addDocumentListener(new thisDocumentListener());
		minmaxValuePanel.add (numPrefixField);

        minFileNumberLabel = new JLabel("  Min:");
        minFileNumberLabel.setHorizontalAlignment (JLabel.RIGHT);
        minmaxValuePanel.add(minFileNumberLabel);

        minFileNumberModel = new SpinnerNumberModel(new Long(minFileNumber),new Long(0), null, new Long(1));
        minFileNumberSpinner = new JSpinner(minFileNumberModel);
        JSpinner.NumberEditor minFileNumberEditor = new JSpinner.NumberEditor( minFileNumberSpinner, "#" );
        minFileNumberSpinner.setEditor( minFileNumberEditor );

        minFileNumberSpinner.setPreferredSize (new Dimension(80,subpanelHeight));
        minFileNumberSpinner.addChangeListener (this);
        minFileNumberSpinner.addMouseWheelListener(this);
        minmaxValuePanel.add(minFileNumberSpinner);

        maxFileNumberLabel = new JLabel("  Max:");
        maxFileNumberLabel.setHorizontalAlignment (JLabel.RIGHT);
        minmaxValuePanel.add(maxFileNumberLabel);

        maxFileNumberModel = new SpinnerNumberModel(new Long(maxFileNumber),new Long(0), null, new Long(1));
        maxFileNumberSpinner = new JSpinner(maxFileNumberModel);
        JSpinner.NumberEditor maxFileNumberEditor = new JSpinner.NumberEditor( maxFileNumberSpinner, "#" );
        maxFileNumberSpinner.setEditor( maxFileNumberEditor );

        maxFileNumberSpinner.setPreferredSize (new Dimension(80,subpanelHeight));
        maxFileNumberSpinner.addChangeListener (this);
        maxFileNumberSpinner.addMouseWheelListener(this);
        minmaxValuePanel.add(maxFileNumberSpinner);

        SpringUtil.makeCompactGrid (minmaxValuePanel, 1, minmaxValuePanel.getComponentCount(), 6,6,6,6);
        mainPanel.add(minmaxValuePanel);

        validFilesLabel = new JLabel ("");
        validFilesLabel.setFont(b12);
		validFilesLabel.setPreferredSize (new Dimension(20,subpanelHeight));
        mainPanel.add (validFilesLabel);





        JLabel headerLabel = new JLabel ("Dark ExpTime Scaling:");
        headerLabel.setFont(p12);
        headerLabel.setBorder(BorderFactory.createEmptyBorder(0, 0, 0, 0));
        headerLabel.setHorizontalAlignment (JTextField.RIGHT);
		mainPanel.add (headerLabel);

        JPanel headerLeftPanel = new JPanel(new SpringLayout());
        headerLeftPanel.setPreferredSize (new Dimension(subpanelWidth,subpanelHeight));

		scaleExpTimeBox = new JCheckBox("Use",scaleExpTime);
        scaleExpTimeBox.setPreferredSize (new Dimension(subpanelWidth,boxHeight));
        scaleExpTimeBox.setToolTipText("<html>"+"Scale the dark image pixel values by the"+"<br>"+
                            "ratio of the science to the dark exposure times."+"<br>"+
                            "Requires the use of a master bias and master dark."+"</html>");
		scaleExpTimeBox.addItemListener (this);
		headerLeftPanel.add (scaleExpTimeBox);

        SpringUtil.makeCompactGrid (headerLeftPanel, 1, headerLeftPanel.getComponentCount(), 6,6,6,6);
        mainPanel.add (headerLeftPanel);

        JPanel headerMiddlePanel = new JPanel(new SpringLayout());
        headerMiddlePanel.setBorder(BorderFactory.createEmptyBorder(10, 0, 0, 0));

        JLabel blankSpace10 = new JLabel ("");
        blankSpace10.setFont(p12);
        blankSpace10.setPreferredSize (new Dimension(110,20));
        if (IJ.isLinux()) blankSpace10.setPreferredSize (new Dimension(120,20));
        blankSpace10.setHorizontalAlignment (JTextField.LEFT);
		headerMiddlePanel.add (blankSpace10);

        JLabel headerMiddleLabel1 = new JLabel ("Relative Directory");
        headerMiddleLabel1.setFont(p12);
        headerMiddleLabel1.setPreferredSize (new Dimension(145,20));
        if (IJ.isLinux()) headerMiddleLabel1.setPreferredSize (new Dimension(142,20));
        headerMiddleLabel1.setHorizontalAlignment (JTextField.LEFT);
		headerMiddlePanel.add (headerMiddleLabel1);

        JLabel headerMiddleLabel2 = new JLabel ("Filename Pattern");
        headerMiddleLabel2.setFont(p12);
        headerMiddleLabel2.setPreferredSize (new Dimension(145,20));
        headerMiddleLabel2.setHorizontalAlignment (JTextField.LEFT);
		headerMiddlePanel.add (headerMiddleLabel2);

        JLabel headerMiddleLabel3 = new JLabel ("#");
        headerMiddleLabel3.setFont(p12);
        headerMiddleLabel3.setPreferredSize (new Dimension(20,20));
        headerMiddleLabel3.setHorizontalAlignment (JTextField.CENTER);
		headerMiddlePanel.add (headerMiddleLabel3);

        SpringUtil.makeCompactGrid (headerMiddlePanel, 1, headerMiddlePanel.getComponentCount(), 0,0,0,0);
        mainPanel.add (headerMiddlePanel);

        JPanel headerRightPanel = new JPanel(new SpringLayout());
        headerRightPanel.setBorder(BorderFactory.createEmptyBorder(10, 7, 0, 0));

        JLabel headerRightLabel = new JLabel ("Master Filename");
        headerRightLabel.setFont(p12);
        headerRightLabel.setPreferredSize (new Dimension(130,20));
        headerRightLabel.setHorizontalAlignment (JTextField.LEFT);
		headerRightPanel.add (headerRightLabel);

        JLabel headerRightLabel2 = new JLabel ("");
		headerRightPanel.add (headerRightLabel2);

        JLabel headerRightLabel3 = new JLabel ("#");
        headerRightLabel3.setFont(p12);
        headerRightLabel3.setPreferredSize (new Dimension(20,20));
        headerRightLabel3.setHorizontalAlignment (JTextField.RIGHT);
		headerRightPanel.add (headerRightLabel3);

        SpringUtil.makeCompactGrid (headerRightPanel, 1, headerRightPanel.getComponentCount(), 0,0,0,0);
        mainPanel.add (headerRightPanel);


        JLabel biasLabel = new JLabel ("Master Bias: ");
        biasLabel.setFont(p12);
        biasLabel.setHorizontalAlignment (JTextField.RIGHT);
		mainPanel.add (biasLabel);

        JPanel masterBiasLeftPanel = new JPanel(new SpringLayout());
        masterBiasLeftPanel.setPreferredSize (new Dimension(subpanelWidth,subpanelHeight));

		useBiasBox = new JCheckBox("Use",useBias);
        useBiasBox.setPreferredSize (new Dimension(boxWidthLeft,boxHeight));
        useBiasBox.setToolTipText("Use the master bias file for master dark and flat creation and science data reduction");
		useBiasBox.addItemListener (this);
		masterBiasLeftPanel.add (useBiasBox);

        createBiasBox = new JCheckBox("Create",createBias);
        createBiasBox.setPreferredSize (new Dimension(boxWidthRight,boxHeight));
        createBiasBox.setToolTipText("Force a new master bias file to be created");
		createBiasBox.addItemListener (this);
        masterBiasLeftPanel.add(createBiasBox);
        SpringUtil.makeCompactGrid (masterBiasLeftPanel, 1, masterBiasLeftPanel.getComponentCount(), 6,6,6,6);
        mainPanel.add (masterBiasLeftPanel);


        JPanel masterBiasMiddlePanel = new JPanel(new SpringLayout());

        biasRadioGroup = new ButtonGroup ();

        biasAverageRadio = new JRadioButton("ave");
        if (!biasMedian) biasAverageRadio.setSelected(true);
        else biasAverageRadio.setSelected(false);
        masterBiasMiddlePanel.add (biasAverageRadio);
        biasMedianRadio = new JRadioButton("med");
        if (biasMedian) biasMedianRadio.setSelected(true);
        else biasMedianRadio.setSelected(false);
        masterBiasMiddlePanel.add (biasMedianRadio);
        biasRadioGroup.add(biasAverageRadio);
        biasRadioGroup.add(biasMedianRadio);
        biasAverageRadio.addActionListener(this);
        biasMedianRadio.addActionListener(this);



		biasDirField = new JTextField (biasDir);
		biasDirField.setPreferredSize (new Dimension(80,subpanelHeight));
		biasDirField.setHorizontalAlignment (JTextField.LEFT);
        biasDirField.setToolTipText("<html>"+"This string defines the directory relative to the primary directory"+"<br>"+
                                             "where raw bias images and the master bias are located (or will be written)."+"<br>"+
                                             "Leave blank to use the primary directory (.."+slash+" notation is supported)." +"</html>");
        biasDirField.getDocument().addDocumentListener(new thisDocumentListener());
		masterBiasMiddlePanel.add (biasDirField);

		biasBaseField = new JTextField (biasBase);
		biasBaseField.setPreferredSize (new Dimension(80,subpanelHeight));
        biasBaseField.setToolTipText("Enter a bias filename pattern using ? and * as wildcard characters");
		biasBaseField.setHorizontalAlignment (JTextField.LEFT);
        biasBaseField.getDocument().addDocumentListener(new thisDocumentListener());
		masterBiasMiddlePanel.add (biasBaseField);

        validBiasFilesLabel = new JLabel ("0");
        validBiasFilesLabel.setFont(b12);
        validBiasFilesLabel.setToolTipText("The number of raw bias files in the directory matching the filename pattern");
        validBiasFilesLabel.setHorizontalAlignment (JLabel.RIGHT);
		validBiasFilesLabel.setPreferredSize (new Dimension(30,subpanelHeight));
        masterBiasMiddlePanel.add (validBiasFilesLabel);

        SpringUtil.makeCompactGrid (masterBiasMiddlePanel, 1, masterBiasMiddlePanel.getComponentCount(), 6,6,6,6);
        mainPanel.add (masterBiasMiddlePanel);


        JPanel masterBiasRightPanel = new JPanel(new SpringLayout());

		biasMasterField = new JTextField (biasMaster);
		biasMasterField.setPreferredSize (new Dimension(60,subpanelHeight));
        biasMasterField.setToolTipText("Enter the master bias filename (including filetype)");
		biasMasterField.setHorizontalAlignment (JTextField.LEFT);
        biasMasterField.getDocument().addDocumentListener(new thisDocumentListener());
		masterBiasRightPanel.add (biasMasterField);
//        JLabel biasMasterLabel = new JLabel(".fits");
//        biasMasterLabel.setFont(b12);
//        masterBiasRightPanel.add (biasMasterLabel);

        validMasterBiasFilesLabel = new JLabel ("0");
        validMasterBiasFilesLabel.setFont(b12);
        validMasterBiasFilesLabel.setToolTipText("The number of master bias files matching the filename (0 or 1)");
        validMasterBiasFilesLabel.setHorizontalAlignment (JLabel.CENTER);
		validMasterBiasFilesLabel.setPreferredSize (new Dimension(20,subpanelHeight));
        masterBiasRightPanel.add (validMasterBiasFilesLabel);
        
        SpringUtil.makeCompactGrid (masterBiasRightPanel, 1, masterBiasRightPanel.getComponentCount(), 6,6,6,6);
        mainPanel.add (masterBiasRightPanel);




        JLabel darkLabel = new JLabel ("Master Dark: ");
        darkLabel.setFont(p12);
        darkLabel.setHorizontalAlignment (JTextField.RIGHT);
		mainPanel.add (darkLabel);

        JPanel masterDarkLeftPanel = new JPanel(new SpringLayout());
        masterDarkLeftPanel.setPreferredSize (new Dimension(subpanelWidth,subpanelHeight));

		useDarkBox = new JCheckBox("Use",useDark);
        useDarkBox.setPreferredSize (new Dimension(boxWidthLeft,boxHeight));
        useDarkBox.setToolTipText("Use the master dark file for master flat creation and science data reduction");
		useDarkBox.addItemListener (this);
		masterDarkLeftPanel.add (useDarkBox);

        createDarkBox = new JCheckBox("Create",createDark);
        createDarkBox.setPreferredSize (new Dimension(boxWidthRight,boxHeight));
        createDarkBox.setToolTipText("Force a new master dark file to be created");
		createDarkBox.addItemListener (this);
        masterDarkLeftPanel.add(createDarkBox);

        SpringUtil.makeCompactGrid (masterDarkLeftPanel, 1, masterDarkLeftPanel.getComponentCount(), 6,6,6,6);
        mainPanel.add (masterDarkLeftPanel);



        JPanel masterDarkMiddlePanel = new JPanel(new SpringLayout());

        darkRadioGroup = new ButtonGroup ();

        darkAverageRadio = new JRadioButton("ave");
        if (!darkMedian) darkAverageRadio.setSelected(true);
        else darkAverageRadio.setSelected(false);
        masterDarkMiddlePanel.add (darkAverageRadio);
        darkMedianRadio = new JRadioButton("med");
        if (darkMedian) darkMedianRadio.setSelected(true);
        else darkMedianRadio.setSelected(false);
        masterDarkMiddlePanel.add (darkMedianRadio);
        darkRadioGroup.add(darkAverageRadio);
        darkRadioGroup.add(darkMedianRadio);
        darkAverageRadio.addActionListener(this);
        darkMedianRadio.addActionListener(this);

		darkDirField = new JTextField (darkDir);
		darkDirField.setPreferredSize (new Dimension(80,subpanelHeight));
        darkDirField.setToolTipText("<html>"+"This string defines the directory relative to the primary directory"+"<br>"+
                                             "where raw dark images and the master dark are located (or will be written)."+"<br>"+
                                             "Leave blank to use the primary directory (.."+slash+" notation is supported)." +"</html>");
		darkDirField.setHorizontalAlignment (JTextField.LEFT);
        darkDirField.getDocument().addDocumentListener(new thisDocumentListener());
		masterDarkMiddlePanel.add (darkDirField);

		darkBaseField = new JTextField (darkBase);
		darkBaseField.setPreferredSize (new Dimension(80,subpanelHeight));
        darkBaseField.setToolTipText("Enter a dark filename pattern using ? and * as wildcard characters");
		darkBaseField.setHorizontalAlignment (JTextField.LEFT);
        darkBaseField.getDocument().addDocumentListener(new thisDocumentListener());
		masterDarkMiddlePanel.add (darkBaseField);

        validDarkFilesLabel = new JLabel ("0");
        validDarkFilesLabel.setFont(b12);
        validDarkFilesLabel.setToolTipText("The number of raw dark files in the directory matching the filename pattern");
        validDarkFilesLabel.setHorizontalAlignment (JLabel.RIGHT);
		validDarkFilesLabel.setPreferredSize (new Dimension(30,subpanelHeight));
        masterDarkMiddlePanel.add (validDarkFilesLabel);

        SpringUtil.makeCompactGrid (masterDarkMiddlePanel, 1, masterDarkMiddlePanel.getComponentCount(), 6,6,6,6);
        mainPanel.add (masterDarkMiddlePanel);



        JPanel masterDarkRightPanel = new JPanel(new SpringLayout());

		darkMasterField = new JTextField (darkMaster);
		darkMasterField.setPreferredSize (new Dimension(60,subpanelHeight));
        darkMasterField.setToolTipText("Enter the master dark filename (including filetype)");
		darkMasterField.setHorizontalAlignment (JTextField.LEFT);
        darkMasterField.getDocument().addDocumentListener(new thisDocumentListener());
		masterDarkRightPanel.add (darkMasterField);
//        JLabel darkMasterLabel = new JLabel(".fits");
//        darkMasterLabel.setFont(b12);
//        masterDarkRightPanel.add (darkMasterLabel);

        validMasterDarkFilesLabel = new JLabel ("0");
        validMasterDarkFilesLabel.setFont(b12);
        validMasterDarkFilesLabel.setToolTipText("The number of master dark files matching the filename (0 or 1)");
        validMasterDarkFilesLabel.setHorizontalAlignment (JLabel.CENTER);
		validMasterDarkFilesLabel.setPreferredSize (new Dimension(20,subpanelHeight));
        masterDarkRightPanel.add (validMasterDarkFilesLabel);

        SpringUtil.makeCompactGrid (masterDarkRightPanel, 1, masterDarkRightPanel.getComponentCount(), 6,6,6,6);
        mainPanel.add (masterDarkRightPanel);





        JLabel flatLabel = new JLabel ("Master Flat: ");
        flatLabel.setFont(p12);
        flatLabel.setHorizontalAlignment (JTextField.RIGHT);
		mainPanel.add (flatLabel);

        JPanel masterFlatLeftPanel = new JPanel(new SpringLayout());
        masterFlatLeftPanel.setPreferredSize (new Dimension(subpanelWidth,subpanelHeight));

		useFlatBox = new JCheckBox("Use",useFlat);
        useFlatBox.setPreferredSize (new Dimension(boxWidthLeft,boxHeight));
        useFlatBox.setToolTipText("Use the master flat file for science data reduction");
		useFlatBox.addItemListener (this);
		masterFlatLeftPanel.add (useFlatBox);

        createFlatBox = new JCheckBox("Create",createFlat);
        createFlatBox.setPreferredSize (new Dimension(boxWidthRight,boxHeight));
        createFlatBox.setToolTipText("Force a new master flat file to be created");
		createFlatBox.addItemListener (this);
        masterFlatLeftPanel.add(createFlatBox);
        SpringUtil.makeCompactGrid (masterFlatLeftPanel, 1, masterFlatLeftPanel.getComponentCount(), 6,6,6,6);
        mainPanel.add (masterFlatLeftPanel);



        JPanel masterFlatMiddlePanel = new JPanel(new SpringLayout());

        flatRadioGroup = new ButtonGroup ();

        flatAverageRadio = new JRadioButton("ave");
        if (!flatMedian) flatAverageRadio.setSelected(true);
        else flatAverageRadio.setSelected(false);
        masterFlatMiddlePanel.add (flatAverageRadio);
        flatMedianRadio = new JRadioButton("med");
        if (flatMedian) flatMedianRadio.setSelected(true);
        else flatMedianRadio.setSelected(false);
        masterFlatMiddlePanel.add (flatMedianRadio);
        flatRadioGroup.add(flatAverageRadio);
        flatRadioGroup.add(flatMedianRadio);
        flatAverageRadio.addActionListener(this);
        flatMedianRadio.addActionListener(this);



		flatDirField = new JTextField (flatDir);
		flatDirField.setPreferredSize (new Dimension(80,subpanelHeight));
		flatDirField.setHorizontalAlignment (JTextField.LEFT);
        flatDirField.setToolTipText("<html>"+"This string defines the directory relative to the primary directory"+"<br>"+
                                             "where raw flat field images and the master flat are located (or will be written)."+"<br>"+
                                             "Leave blank to use the primary directory (.."+slash+" notation is supported)." +"</html>");
        flatDirField.getDocument().addDocumentListener(new thisDocumentListener());
		masterFlatMiddlePanel.add (flatDirField);

		flatBaseField = new JTextField (flatBase);
		flatBaseField.setPreferredSize (new Dimension(80,subpanelHeight));
        flatBaseField.setToolTipText("Enter a flat filename pattern using ? and * as wildcard characters");
		flatBaseField.setHorizontalAlignment (JTextField.LEFT);
        flatBaseField.getDocument().addDocumentListener(new thisDocumentListener());
		masterFlatMiddlePanel.add (flatBaseField);

        validFlatFilesLabel = new JLabel ("0");
        validFlatFilesLabel.setFont(b12);
        validFlatFilesLabel.setToolTipText("The number of raw flat files in the directory matching the filename pattern");
        validFlatFilesLabel.setHorizontalAlignment (JLabel.RIGHT);
		validFlatFilesLabel.setPreferredSize (new Dimension(30,subpanelHeight));
        masterFlatMiddlePanel.add (validFlatFilesLabel);

        SpringUtil.makeCompactGrid (masterFlatMiddlePanel, 1, masterFlatMiddlePanel.getComponentCount(), 6,6,6,6);
        mainPanel.add (masterFlatMiddlePanel);



        JPanel masterFlatRightPanel = new JPanel(new SpringLayout());

		flatMasterField = new JTextField (flatMaster);
		flatMasterField.setPreferredSize (new Dimension(60,subpanelHeight));
        flatMasterField.setToolTipText("Enter the master flat filename (including filetype)");
		flatMasterField.setHorizontalAlignment (JTextField.LEFT);
        flatMasterField.getDocument().addDocumentListener(new thisDocumentListener());
		masterFlatRightPanel.add (flatMasterField);
//        JLabel flatMasterLabel = new JLabel(".fits");
//        flatMasterLabel.setFont(b12);
//        masterFlatRightPanel.add (flatMasterLabel);

        validMasterFlatFilesLabel = new JLabel ("0");
        validMasterFlatFilesLabel.setFont(b12);
        validMasterFlatFilesLabel.setToolTipText("The number of master flat files matching the filename (0 or 1)");
        validMasterFlatFilesLabel.setHorizontalAlignment (JLabel.CENTER);
		validMasterFlatFilesLabel.setPreferredSize (new Dimension(20,subpanelHeight));
        masterFlatRightPanel.add (validMasterFlatFilesLabel);

        SpringUtil.makeCompactGrid (masterFlatRightPanel, 1, masterFlatRightPanel.getComponentCount(), 6,6,6,6);
        mainPanel.add (masterFlatRightPanel);


        JLabel linearityLabel = new JLabel ("Non-Linearity Correction: ");
        linearityLabel.setFont(p12);
        linearityLabel.setHorizontalAlignment (JTextField.RIGHT);
		mainPanel.add (linearityLabel);

        JPanel linearityLeftPanel = new JPanel(new SpringLayout());
        linearityLeftPanel.setPreferredSize (new Dimension(subpanelWidth,subpanelHeight));

		useNLCBox = new JCheckBox("Use",useNLC);
        useNLCBox.setPreferredSize (new Dimension(boxWidthLeft,boxHeight));
        useNLCBox.setToolTipText("<html>"+"Compensates for CCD non-linearity on a per pixel basis"+"<br>"+
                                    "using the formula and three values to the right."+"<br>"+
                                    "NLC requires the use of a master bias."+"</html>");
		useNLCBox.addItemListener (this);
		linearityLeftPanel.add (useNLCBox);

        JLabel NLCLabel = new JLabel (" New Value =");
        NLCLabel.setFont(b12);
        NLCLabel.setPreferredSize (new Dimension(boxWidthRight,boxHeight));
        NLCLabel.setHorizontalAlignment (JTextField.LEFT);
		linearityLeftPanel.add (NLCLabel);

        SpringUtil.makeCompactGrid (linearityLeftPanel, 1, linearityLeftPanel.getComponentCount(), 6,6,6,6);
        mainPanel.add (linearityLeftPanel);



        JPanel linearityMiddlePanel = new JPanel(new SpringLayout());
//        JLabel coeffALabel = new JLabel("    A");
//        coeffALabel.setHorizontalAlignment (JLabel.RIGHT);
//        linearityMiddlePanel.add(coeffALabel);

        coeffAModel = new SpinnerNumberModel(new Double(coeffA), null, null, new Double(0.0));
        coeffASpinner = new JSpinner(coeffAModel);
        JSpinner.NumberEditor coeffANumberEditor = new JSpinner.NumberEditor( coeffASpinner, "0.0#######E0" );
        coeffASpinner.setEditor(coeffANumberEditor);
        coeffASpinner.setPreferredSize (new Dimension(80,subpanelHeight));
        coeffASpinner.addChangeListener (this);
//        coeffASpinner.addMouseWheelListener(this);
        linearityMiddlePanel.add(coeffASpinner);

        JLabel coeffBLabel = new JLabel("+");
        coeffBLabel.setHorizontalAlignment (JLabel.LEFT);
        coeffBLabel.setFont(b14);
        linearityMiddlePanel.add(coeffBLabel);

        coeffBModel = new SpinnerNumberModel(new Double(coeffB), null, null, new Double(0.0));
        coeffBSpinner = new JSpinner(coeffBModel);
        JSpinner.NumberEditor coeffBNumberEditor = new JSpinner.NumberEditor( coeffBSpinner, "0.0#######E0" );
        coeffBSpinner.setEditor(coeffBNumberEditor);
        coeffBSpinner.setPreferredSize (new Dimension(80,subpanelHeight));
        coeffBSpinner.addChangeListener (this);
//        coeffBSpinner.addMouseWheelListener(this);
        linearityMiddlePanel.add(coeffBSpinner);

        JLabel coeffCLabel = new JLabel("x+");
        coeffCLabel.setFont(b14);
        coeffCLabel.setHorizontalAlignment (JLabel.RIGHT);
        linearityMiddlePanel.add(coeffCLabel);

        coeffCModel = new SpinnerNumberModel(new Double(coeffC), null, null, new Double(0.0));
        coeffCSpinner = new JSpinner(coeffCModel);
        JSpinner.NumberEditor coeffCNumberEditor = new JSpinner.NumberEditor( coeffCSpinner, "0.0#######E0" );
        coeffCSpinner.setEditor(coeffCNumberEditor);
        coeffCSpinner.setPreferredSize (new Dimension(110,subpanelHeight));
        coeffCSpinner.addChangeListener (this);
//        coeffCSpinner.addMouseWheelListener(this);
        linearityMiddlePanel.add(coeffCSpinner);

        JLabel NLCFinalLabel = new JLabel("x\u00b2 ");
        NLCFinalLabel.setHorizontalAlignment (JLabel.CENTER);
        NLCFinalLabel.setFont(b14);
        linearityMiddlePanel.add(NLCFinalLabel);

        SpringUtil.makeCompactGrid (linearityMiddlePanel, 1, linearityMiddlePanel.getComponentCount(), 6,6,6,6);
        mainPanel.add (linearityMiddlePanel);



        JPanel linearityRightPanel = new JPanel(new SpringLayout());

        JLabel NLCFormulaLabel = new JLabel("");
        NLCFormulaLabel.setHorizontalAlignment (JLabel.CENTER);
        linearityRightPanel.add(NLCFormulaLabel);

        SpringUtil.makeCompactGrid (linearityRightPanel, 1, linearityRightPanel.getComponentCount(), 6,6,6,6);
        mainPanel.add (linearityRightPanel);


        JLabel airmassLabel = new JLabel ("Airmass/HJD/etc Calculations: ");
        airmassLabel.setFont(p12);
        airmassLabel.setHorizontalAlignment (JTextField.RIGHT);
		mainPanel.add (airmassLabel);

        JPanel airmassLeftPanel = new JPanel(new SpringLayout());
        airmassLeftPanel.setPreferredSize (new Dimension(subpanelWidth,subpanelHeight));

		calcAirmassBox = new JCheckBox("Use",calcAirmass);
        calcAirmassBox.setPreferredSize (new Dimension(boxWidthLeft,boxHeight));
        calcAirmassBox.setToolTipText("<html>"+"Calculate the following astrophysical quantities at mid-exposure"+"<br>"+
                                    "if EXPTIME is available in FITS header - start-exposure otherwise:"+"<br>"+
                                    "--Julian Date"+"<br>"+
                                    "--Heliocentric Julian Date"+"<br>"+
                                    "--Object altitude above horizon"+"<br>"+
                                    "--Airmass of observation"+"<br>"+
                                    "The following input is required:"+"<br>"+
                                    "--DATE-OBS from FITS header"+"<br>"+
                                    "--EXPTIME (for mid-exposure calculations)"+"<br>"+
                                    "--RA and DEC of object from fields to the right"+"<br>"+
                                    "--Latitude and Longitude of observatory from fields to the right"+"</html>");
		calcAirmassBox.addItemListener (this);
		airmassLeftPanel.add (calcAirmassBox);

		saveAirmassBox = new JCheckBox("Save",saveAirmass);
        saveAirmassBox.setPreferredSize (new Dimension(boxWidthRight,boxHeight));
        saveAirmassBox.setToolTipText("Save entered and calculated values to FITS header");
		saveAirmassBox.addItemListener (this);
		airmassLeftPanel.add (saveAirmassBox);

        SpringUtil.makeCompactGrid (airmassLeftPanel, 1, airmassLeftPanel.getComponentCount(), 6,6,6,6);
        mainPanel.add (airmassLeftPanel);



        JPanel airmassMiddlePanel = new JPanel(new SpringLayout());

        JTextField raLabelField = new JTextField(" RA:");
//        raLabelField.setFont(p12);
//        raLabelField.setPreferredSize(new Dimension(30,17));
        raLabelField.setHorizontalAlignment(JTextField.RIGHT);
        raLabelField.setBorder(BorderFactory.createEmptyBorder());
        raLabelField.setEditable(false);
        airmassMiddlePanel.add(raLabelField);

        int amWidth = 80;
        raTextField = new JTextField(decToSex(ra, 3, false));
//        raTextField.setFont(p12);
        raTextField.setPreferredSize(new Dimension(amWidth,25));
        raTextField.setHorizontalAlignment(JTextField.LEFT);
//        raTextField.setBorder(BorderFactory.createLineBorder(Color.RED));
        raTextField.setEditable(true);
        raTextField.setToolTipText("Enter the right ascension of the target in either decimal or sexagesimal format");
        raTextField.addActionListener(this);
        raTextField.getDocument().addDocumentListener(new thisDocumentListener());
        airmassMiddlePanel.add(raTextField);

        JTextField decLabelField = new JTextField(" DEC:");
//        decLabelField.setFont(p12);
//        raLabelField.setPreferredSize(new Dimension(30,17));
        decLabelField.setHorizontalAlignment(JTextField.RIGHT);
        decLabelField.setBorder(BorderFactory.createEmptyBorder());
        decLabelField.setEditable(false);
        airmassMiddlePanel.add(decLabelField);

        decTextField = new JTextField(decToSex(dec, 2, true));
//        decTextField.setFont(p12);
        decTextField.setPreferredSize(new Dimension(amWidth,25));
        decTextField.setHorizontalAlignment(JTextField.LEFT);
//        raTextField.setBorder(BorderFactory.createLineBorder(Color.RED));
        decTextField.setEditable(true);
        decTextField.setToolTipText("Enter the declination of the target in either decimal or sexagesimal format");
        decTextField.addActionListener(this);
        decTextField.getDocument().addDocumentListener(new thisDocumentListener());
        airmassMiddlePanel.add(decTextField);

        JTextField latLabelField = new JTextField(" LAT:");
//        latLabelField.setFont(p12);
//        raLabelField.setPreferredSize(new Dimension(30,17));
        latLabelField.setHorizontalAlignment(JTextField.RIGHT);
        latLabelField.setBorder(BorderFactory.createEmptyBorder());
        latLabelField.setEditable(false);
        airmassMiddlePanel.add(latLabelField);

        latTextField = new JTextField(decToSex(lat, 2, false));
//        latTextField.setFont(p12);
        latTextField.setPreferredSize(new Dimension(amWidth,25));
        latTextField.setHorizontalAlignment(JTextField.LEFT);
//        raTextField.setBorder(BorderFactory.createLineBorder(Color.RED));
        latTextField.setEditable(true);
        latTextField.setToolTipText("Enter the latitude of the observatory in either decimal or sexagesimal format");
        latTextField.addActionListener(this);
        latTextField.getDocument().addDocumentListener(new thisDocumentListener());
        airmassMiddlePanel.add(latTextField);       


        SpringUtil.makeCompactGrid (airmassMiddlePanel, 1, airmassMiddlePanel.getComponentCount(), 6,6,6,6);
        mainPanel.add (airmassMiddlePanel);



        JPanel airmassRightPanel = new JPanel(new SpringLayout());

        JTextField lonLabelField = new JTextField("LONG:");
//        lonLabelField.setFont(p12);
        lonLabelField.setPreferredSize(new Dimension(30,25));
        lonLabelField.setHorizontalAlignment(JTextField.RIGHT);
        lonLabelField.setBorder(BorderFactory.createEmptyBorder());
        lonLabelField.setEditable(false);
        airmassRightPanel.add(lonLabelField);

        lonTextField = new JTextField(decToSex(lon, 2, false));
//        lonTextField.setFont(p12);
        lonTextField.setPreferredSize(new Dimension(amWidth,25));
        lonTextField.setHorizontalAlignment(JTextField.LEFT);
//        raTextField.setBorder(BorderFactory.createLineBorder(Color.RED));
        lonTextField.setEditable(true);
        lonTextField.setToolTipText("Enter the longitude of the observatory in either decimal or sexagesimal format");
        lonTextField.addActionListener(this);
        lonTextField.getDocument().addDocumentListener(new thisDocumentListener());
        airmassRightPanel.add(lonTextField);

        JLabel amRightDummyLabel = new JLabel (" ");
        amRightDummyLabel.setPreferredSize(new Dimension(20,25));
        amRightDummyLabel.setFont(p12);
        amRightDummyLabel.setHorizontalAlignment (JTextField.RIGHT);
		airmassRightPanel.add (amRightDummyLabel);  


        SpringUtil.makeCompactGrid (airmassRightPanel, 1, airmassRightPanel.getComponentCount(), 6,6,6,6);
        mainPanel.add (airmassRightPanel);


        JLabel processLabel = new JLabel ("Processed Science Images: ");
        processLabel.setFont(p12);
        processLabel.setHorizontalAlignment (JTextField.RIGHT);
		mainPanel.add (processLabel);

        JPanel processLeftPanel = new JPanel(new SpringLayout());
        processLeftPanel.setPreferredSize (new Dimension(subpanelWidth,subpanelHeight));

        showScienceBox = new JCheckBox("Show",showScience);
        showScienceBox.setPreferredSize (new Dimension(boxWidthLeft,boxHeight));
		showScienceBox.addItemListener (this);
        processLeftPanel.add(showScienceBox);

        saveProcessedDataBox = new JCheckBox("Save",saveProcessedData);
        saveProcessedDataBox.setPreferredSize (new Dimension(boxWidthRight,boxHeight));
		saveProcessedDataBox.addItemListener (this);
        processLeftPanel.add(saveProcessedDataBox);


        SpringUtil.makeCompactGrid (processLeftPanel, 1, processLeftPanel.getComponentCount(), 6,6,6,6);
        mainPanel.add (processLeftPanel);



        JPanel scienceMiddlePanel = new JPanel(new SpringLayout());

        saveRadioGroup = new ButtonGroup ();

        saveIntegerRadio = new JRadioButton("16  ");
        saveIntegerRadio.setToolTipText("select for 16-bit integer output");
        if (!saveFloatingPoint) saveIntegerRadio.setSelected(true);
        else saveIntegerRadio.setSelected(false);
        scienceMiddlePanel.add (saveIntegerRadio);
        saveFloatingPointRadio = new JRadioButton("32  ");
        saveFloatingPointRadio.setToolTipText("select for 32-bit floating point output");
        if (saveFloatingPoint) saveFloatingPointRadio.setSelected(true);
        else saveFloatingPointRadio.setSelected(false);
        scienceMiddlePanel.add (saveFloatingPointRadio);
        saveRadioGroup.add(saveIntegerRadio);
        saveRadioGroup.add(saveFloatingPointRadio);
        saveIntegerRadio.addActionListener(this);
        saveFloatingPointRadio.addActionListener(this);



		saveDirField = new JTextField (saveDir);
		saveDirField.setPreferredSize (new Dimension(95,subpanelHeight));
        saveDirField.setToolTipText("<html>"+"Reduced science images are stored in this subdirectory."+"<br>"+
                                             "Leave blank to save in primary directory." +"</html>");
		saveDirField.setHorizontalAlignment (JTextField.LEFT);
        saveDirField.getDocument().addDocumentListener(new thisDocumentListener());
		scienceMiddlePanel.add (saveDirField);

        JLabel suffixLabel = new JLabel("  Suffix");
		suffixLabel.setHorizontalAlignment (JTextField.RIGHT);
        scienceMiddlePanel.add (suffixLabel);

		saveSuffixField = new JTextField (saveSuffix);
		saveSuffixField.setPreferredSize (new Dimension(30,subpanelHeight));
        saveSuffixField.setToolTipText("<html>"+"This string is appended to all science image"+"<br>"+
                                                "output filenames (before the filetype)."+"<br>"+
                                                "Leave blank for no suffix." +"</html>");
		saveSuffixField.setHorizontalAlignment (JTextField.LEFT);
        saveSuffixField.getDocument().addDocumentListener(new thisDocumentListener());
		scienceMiddlePanel.add (saveSuffixField);

        SpringUtil.makeCompactGrid (scienceMiddlePanel, 1, scienceMiddlePanel.getComponentCount(), 6,6,6,6);
        mainPanel.add (scienceMiddlePanel);



        JPanel scienceRightPanel = new JPanel(new SpringLayout());

        JLabel saveFormatLabel = new JLabel("Format");
		saveFormatLabel.setHorizontalAlignment (JTextField.RIGHT);
        scienceRightPanel.add (saveFormatLabel);

		saveFormatField = new JTextField (saveFormat);
		saveFormatField.setPreferredSize (new Dimension(40,subpanelHeight));
        saveFormatField.setToolTipText("<html>"+"Enter the file save format:"+"<br>"+
                                    "-Leave blank to use the input format"+"<br>"+
                                    "-FITS format valid designations are:"+"<br>"+
                                    "--- .fits .FITS .fit, .FIT .fts .FTS"+"<br>"+
                                    "-Other formats include all native ImageJ formats such as:"+"<br>"+
                                    "--- .tif .jpg .gif .bmp .png .raw .avi .zip"+"</html>");
		saveFormatField.setHorizontalAlignment (JTextField.LEFT);
        saveFormatField.getDocument().addDocumentListener(new thisDocumentListener());
		scienceRightPanel.add (saveFormatField);

        SpringUtil.makeCompactGrid (scienceRightPanel, 1, scienceRightPanel.getComponentCount(), 6,6,6,6);
        mainPanel.add (scienceRightPanel);

        JLabel beepLabel = new JLabel (" ");
        beepLabel.setFont(p12);
        beepLabel.setHorizontalAlignment (JTextField.RIGHT);
		mainPanel.add (beepLabel);

        JPanel leftBeep = new JPanel(new SpringLayout());
        leftBeep.setPreferredSize (new Dimension(subpanelWidth,subpanelHeight));

		useBeepBox = new JCheckBox("Beep",useBeep);
        useBeepBox.setPreferredSize (new Dimension(boxWidthLeft,boxHeight));
		useBeepBox.addItemListener (this);
		leftBeep.add (useBeepBox);

        preMacro1AutoLevelBox = new JCheckBox("Auto Level",preMacro1AutoLevel);
        preMacro1AutoLevelBox.setPreferredSize (new Dimension(boxWidthRight,boxHeight));
		preMacro1AutoLevelBox.addItemListener (this);
        leftBeep.add(preMacro1AutoLevelBox);

        SpringUtil.makeCompactGrid (leftBeep, 1, leftBeep.getComponentCount(), 6,6,6,6);
        mainPanel.add(leftBeep);

        JPanel optionPanel = new JPanel(new SpringLayout());
        showLogBox = new JCheckBox("Show Log",showLog);
//        showLogBox.setPreferredSize (new Dimension(boxWidthRight,boxHeight));
		showLogBox.addItemListener (this);
        optionPanel.add(showLogBox);

		showMastersBox = new JCheckBox("Show Master Images",showMasters);
//        showMastersBox.setPreferredSize (new Dimension(boxWidthRight,boxHeight));
        showMastersBox.setToolTipText("Show the selected master bias, dark, and flat images");
		showMastersBox.addItemListener (this);
		optionPanel.add (showMastersBox);

		useGradientRemovalBox = new JCheckBox("Gradient Removal", useGradientRemoval);
//        useGradientRemovalBox.setPreferredSize (new Dimension(boxWidthRight,boxHeight));
        useGradientRemovalBox.setToolTipText("Remove average gradient from flats when creating master flat-field image");
		useGradientRemovalBox.addItemListener (this);
		optionPanel.add (useGradientRemovalBox);



        SpringUtil.makeCompactGrid (optionPanel, 1, optionPanel.getComponentCount(), 6,6,6,6);
        mainPanel.add(optionPanel);

		useCosmicRemovalBox = new JCheckBox("Cosmic Ray Removal", useCosmicRemoval);
//        useCosmicRemovalBox.setPreferredSize (new Dimension(boxWidthRight,boxHeight));
        useCosmicRemovalBox.setToolTipText("Remove hot pixels and cosmic ray hits from science image after all other data processing");
		useCosmicRemovalBox.addItemListener (this);
		mainPanel.add (useCosmicRemovalBox);

//        JLabel blankSpace2 = new JLabel ("");
//		blankSpace2.setPreferredSize (new Dimension(20,subpanelHeight));
//        mainPanel.add (blankSpace2);


        JLabel macro1Label = new JLabel ("Macro 1: ");
        macro1Label.setFont(p12);
        macro1Label.setHorizontalAlignment (JTextField.RIGHT);
		mainPanel.add (macro1Label);

        JPanel leftMacro1 = new JPanel(new SpringLayout());
        leftMacro1.setPreferredSize (new Dimension(subpanelWidth,subpanelHeight));

		runPreMacroBox = new JCheckBox("Use",runPreMacro);
        runPreMacroBox.setPreferredSize (new Dimension(boxWidthLeft,boxHeight));
		runPreMacroBox.addItemListener (this);
        leftMacro1.add(runPreMacroBox);

		postMacro1AutoLevelBox = new JCheckBox("Auto Level",postMacro1AutoLevel);
        postMacro1AutoLevelBox.setPreferredSize (new Dimension(boxWidthRight,boxHeight));
		postMacro1AutoLevelBox.addItemListener (this);
        leftMacro1.add(postMacro1AutoLevelBox);

        SpringUtil.makeCompactGrid (leftMacro1, 1, leftMacro1.getComponentCount(), 6,6,6,6);
        mainPanel.add (leftMacro1);

		preMacroText = new JTextField (preMacroPath);
        preMacroText.getDocument().addDocumentListener(new thisDocumentListener());
		preMacroText.setPreferredSize (new Dimension(300,subpanelHeight));
		preMacroText.setHorizontalAlignment (JTextField.LEFT);
		mainPanel.add (preMacroText);

		preMacroButton = new JButton ("Browse");
		preMacroButton.addActionListener (this);
		mainPanel.add (preMacroButton);


        JLabel macro2Label = new JLabel ("Macro 2: ");
        macro2Label.setFont(p12);
        macro2Label.setHorizontalAlignment (JTextField.RIGHT);
		mainPanel.add (macro2Label);

        JPanel leftMacro2 = new JPanel(new SpringLayout());
        leftMacro2.setPreferredSize (new Dimension(subpanelWidth,subpanelHeight));

		runPostMacroBox = new JCheckBox("Use",runPostMacro);
        runPostMacroBox.setPreferredSize (new Dimension(boxWidthLeft,boxHeight));
		runPostMacroBox.addItemListener (this);
        leftMacro2.add(runPostMacroBox);

		postMacro2AutoLevelBox = new JCheckBox("Auto Level",postMacro2AutoLevel);
        postMacro2AutoLevelBox.setPreferredSize (new Dimension(boxWidthRight,boxHeight));
		postMacro2AutoLevelBox.addItemListener (this);
        leftMacro2.add(postMacro2AutoLevelBox);

        SpringUtil.makeCompactGrid (leftMacro2, 1, leftMacro2.getComponentCount(), 6,6,6,6);
        mainPanel.add (leftMacro2);

		postMacroText = new JTextField (postMacroPath);
        postMacroText.getDocument().addDocumentListener(new thisDocumentListener());
		postMacroText.setPreferredSize (new Dimension(300,subpanelHeight));
		postMacroText.setHorizontalAlignment (JTextField.LEFT);
		mainPanel.add (postMacroText);

		postMacroButton = new JButton ("Browse");
		postMacroButton.addActionListener (this);
		mainPanel.add (postMacroButton);

                
 

		statusLabel = new JLabel ("");
        statusLabel.setFont(b11);
        updateStatusLabel();
        statusLabel.setPreferredSize (new Dimension(170,30));
        statusLabel.setHorizontalAlignment (JTextField.RIGHT);
		mainPanel.add (statusLabel);

        JPanel onlyNewPanel = new JPanel(new SpringLayout());
        onlyNewPanel.setPreferredSize (new Dimension(subpanelWidth,subpanelHeight));

		onlyNewBox = new JCheckBox("New Arrivals Only", onlyNew);
        onlyNewBox.setPreferredSize (new Dimension(subpanelWidth,30));
        onlyNewBox.setToolTipText("Process only files that arrive after pressing the start button.");

		onlyNewBox.addItemListener (this);
        onlyNewPanel.add(onlyNewBox);

        SpringUtil.makeCompactGrid (onlyNewPanel, 1, onlyNewPanel.getComponentCount(), 6,6,6,6);
		mainPanel.add (onlyNewPanel);

		JPanel controlPanel = new JPanel(new SpringLayout());
		startButton = new JButton ("START");
        startButton.setFont(b12);
        startButton.setPreferredSize(new Dimension(150, 30));
		startButton.addActionListener (this);
		controlPanel.add (startButton);

		stopButton = new JButton ("STOP");
		stopButton.addActionListener (this);
        stopButton.setFont(b12);
        stopButton.setPreferredSize(new Dimension(100, 30));
		controlPanel.add (stopButton);

		JLabel blankSpace9 = new JLabel ("");
        blankSpace9.setPreferredSize (new Dimension(20,subpanelHeight));
        controlPanel.add (blankSpace9);

        clearButton = new JButton("RESET");
        clearButton.setPreferredSize(new Dimension(80, 30));
        clearButton.setFont(b12);
        clearButton.addActionListener (this);
        clearButton.setToolTipText("<html>"+"Resets the file processing queue."+"<br>"+
                                   "Pressing START after RESET causes all files to be processed again (if New Arrivals Only is UNchecked)."+"</html>");
        controlPanel.add(clearButton);

		JLabel blankSpace8 = new JLabel ("");
//		blankSpace2.setPreferredSize (new Dimension(00,subpanelHeight));
        controlPanel.add (blankSpace8);

        SpringUtil.makeCompactGrid (controlPanel, 1, controlPanel.getComponentCount(), 6,6,6,6);


		mainPanel.add(controlPanel);

        JPanel pollingPanel = new JPanel(new SpringLayout());
        pollingIntervalLabel = new JLabel("Polling Interval:");
        pollingIntervalLabel.setToolTipText("<html>"+"Set polling interval to 0 to disable polling."+"<br>"+
                                   "For polling intervals > 0, the primary directory will be rescanned at that interval (in secs)."+"</html>");
        pollingIntervalLabel.setHorizontalAlignment (JLabel.RIGHT);
        pollingPanel.add(pollingIntervalLabel);

        pollingIntervalModel = new SpinnerNumberModel(new Integer(pollingInterval),new Integer(0), null, new Integer(1));
        pollingIntervalSpinner = new JSpinner(pollingIntervalModel);
        pollingIntervalSpinner.setPreferredSize (new Dimension(50,subpanelHeight));
        pollingIntervalSpinner.setToolTipText("<html>"+"Set polling interval to 0 to disable polling."+"<br>"+
                                   "For polling intervals > 0, the primary directory will be rescanned at that interval (in secs)."+"</html>");
        pollingIntervalSpinner.addChangeListener (this);
        pollingIntervalSpinner.addMouseWheelListener(this);
     
        pollingPanel.add(pollingIntervalSpinner);
        SpringUtil.makeCompactGrid (pollingPanel, 1, pollingPanel.getComponentCount(), 6,6,6,6);
        mainPanel.add(pollingPanel);



		SpringUtil.makeCompactGrid (mainPanel,mainPanel.getComponentCount()/4, 4, 6,6,6,6);
		dialogFrame.add (mainPanel);
		dialogFrame.pack();
		dialogFrame.setResizable (false);
        dialogFrame.setLocation(dialogFrameLocationX, dialogFrameLocationY);
		dialogFrame.setVisible (true);

        countValidFiles();
		}

    void saveAndClose()
            {
            if (timer != null) timer.cancel();
            if (task != null) task.cancel();

            savePrefs();
            
            dialogFrame.dispose();

            }

    public void stateChanged(ChangeEvent ev)
        {
        if (ev.getSource() == minFileNumberSpinner)
            {
            minFileNumber = ((Long)minFileNumberSpinner.getValue()).longValue();
            Prefs.set("watcher.minFileNumber", minFileNumber);
            countValidFiles();
            }
        else if (ev.getSource() == maxFileNumberSpinner)
            {
            maxFileNumber = ((Long)maxFileNumberSpinner.getValue()).longValue();
            Prefs.set("watcher.maxFileNumber", maxFileNumber);
            countValidFiles();
            }
        else if (ev.getSource() == pollingIntervalSpinner)
            {
            pollingInterval = ((Integer)pollingIntervalSpinner.getValue()).intValue();
            Prefs.set("watcher.pollingInterval", pollingInterval);
            }
        else if (ev.getSource() == coeffASpinner)
            {
            coeffA = ((Double)coeffASpinner.getValue()).doubleValue();
            Prefs.set("watcher.coeffA", coeffA);
            }
        else if (ev.getSource() == coeffBSpinner)
            {
            coeffB = ((Double)coeffBSpinner.getValue()).doubleValue();
            Prefs.set("watcher.coeffB", coeffB);
            }
        else if (ev.getSource() == coeffCSpinner)
            {
            coeffC = ((Double)coeffCSpinner.getValue()).doubleValue();
            Prefs.set("watcher.coeffC", coeffC);
            }
        }


    public void mouseWheelMoved( MouseWheelEvent e )
        {
        if (e.getSource() == minFileNumberSpinner)
            {
            int newValue =  new Integer(((Integer)minFileNumberSpinner.getValue()).intValue() - e.getWheelRotation());
            if (newValue >= 0) minFileNumberSpinner.setValue(newValue);
            }
        else if (e.getSource() == maxFileNumberSpinner)
            {
            int newValue =  new Integer(((Integer)maxFileNumberSpinner.getValue()).intValue() - e.getWheelRotation());
            if (newValue >= 0) maxFileNumberSpinner.setValue(newValue);
            }
        else if (e.getSource() == pollingIntervalSpinner)
            {
            int newValue =  new Integer(((Integer)pollingIntervalSpinner.getValue()).intValue() - e.getWheelRotation());
            if (newValue > 0) pollingIntervalSpinner.setValue(newValue);
            }
        else if (e.getSource() == coeffASpinner)
            {
            coeffASpinner.setValue(((Double)coeffASpinner.getValue()).doubleValue()
                    - e.getWheelRotation()*0.0 );
            }
        else if (e.getSource() == coeffBSpinner)
            {
            coeffBSpinner.setValue(((Double)coeffBSpinner.getValue()).doubleValue()
                    - e.getWheelRotation()*0.0 );
            }
        else if (e.getSource() == coeffCSpinner)
            {
            coeffCSpinner.setValue(((Double)coeffCSpinner.getValue()).doubleValue()
                    - e.getWheelRotation()*0.0 );
            }
        }


	public void itemStateChanged (ItemEvent e)
		{
		Object source = e.getItemSelectable();
		if (e.getStateChange() == ItemEvent.SELECTED)
			{
			if (source == onlyNewBox)
				onlyNew = true;
			else if (source == runPreMacroBox)
				runPreMacro = true;
			else if (source == runPostMacroBox)
				runPostMacro = true;
            else if (source == preMacro1AutoLevelBox)
                preMacro1AutoLevel = true;
            else if (source == postMacro1AutoLevelBox)
                postMacro1AutoLevel = true;
            else if (source == postMacro2AutoLevelBox)
                postMacro2AutoLevel = true;
			else if (source == useBeepBox)
				useBeep = true;
            else if (source == sortNumericallyBox)
				sortNumerically = true;
            else if (source == updateExcludeBox)
                {
				updateExclude = true;
                if (filenamePattern.lastIndexOf('*') >= 0)
                    {
                    numPrefix = filenamePattern.substring(0, filenamePattern.lastIndexOf('*'));
                    numPrefixField.setText(numPrefix);
                    }
                else
                    {
                    numPrefix = filenamePattern;
                    numPrefixField.setText(numPrefix);
                    }
                }
            else if (source == useBiasBox)
				useBias = true;
            else if (source == createBiasBox)
				createBias = true;
            else if (source == useDarkBox)
				useDark = true;
            else if (source == createDarkBox)
				createDark = true;
            else if (source == useFlatBox)
				useFlat = true;
            else if (source == createFlatBox)
				createFlat = true;
            else if (source == useNLCBox)
				useNLC = true;
            else if (source == scaleExpTimeBox)
				scaleExpTime = true;
            else if (source == saveProcessedDataBox)
				saveProcessedData = true;
            else if (source == showMastersBox)
				showMasters = true;
            else if (source == showScienceBox)
				showScience = true;
            else if (source == showLogBox)
				showLog = true;
            else if (source == useGradientRemovalBox)
				useGradientRemoval = true;
            else if (source == useCosmicRemovalBox)
				useCosmicRemoval = true;
            else if (source == calcAirmassBox)
				calcAirmass = true;
            else if (source == saveAirmassBox)
				saveAirmass = true;


			}
		else if (e.getStateChange() == ItemEvent.DESELECTED)
			{
			if (source == onlyNewBox)
				onlyNew = false;
			else if (source == runPreMacroBox)
				runPreMacro = false;
			else if (source == runPostMacroBox)
				runPostMacro = false;          
            else if (source == preMacro1AutoLevelBox)
                preMacro1AutoLevel = false;
            else if (source == postMacro1AutoLevelBox)
                postMacro1AutoLevel = false;            
            else if (source == postMacro2AutoLevelBox)
                postMacro2AutoLevel = false;
			else if (source == useBeepBox)
				useBeep = false;
            else if (source == sortNumericallyBox)
				sortNumerically = false;
            else if (source == updateExcludeBox)
				updateExclude = false;
            else if (source == useBiasBox)
				useBias = false;
            else if (source == createBiasBox)
				createBias = false;
            else if (source == useDarkBox)
				useDark = false;
            else if (source == createDarkBox)
				createDark = false;
            else if (source == useFlatBox)
				useFlat = false;
            else if (source == createFlatBox)
				createFlat = false;
            else if (source == useNLCBox)
				useNLC = false;
            else if (source == scaleExpTimeBox)
				scaleExpTime = false;
            else if (source == saveProcessedDataBox)
				saveProcessedData = false;
            else if (source == showMastersBox)
				showMasters = false;
            else if (source == showScienceBox)
				showScience = false;
            else if (source == showLogBox)
				showLog = false;
            else if (source == useGradientRemovalBox)
				useGradientRemoval = false;
            else if (source == useCosmicRemovalBox)
				useCosmicRemoval = false;
            else if (source == calcAirmassBox)
				calcAirmass = false;
            else if (source == saveAirmassBox)
				saveAirmass = false;
			}
		}

	/**
	 * Response to buttons being pushed.
	 */
	public void actionPerformed (ActionEvent ae)
		{
        Object source = ae.getSource();
        getTextFields();
        countValidFiles();
        if (source == saveIntegerRadio)
                saveFloatingPoint = false;
        else if (source == saveFloatingPointRadio)
                saveFloatingPoint = true;
        else if (source == flatAverageRadio)
                flatMedian = false;
        else if (source == flatMedianRadio)
                flatMedian = true;
        else if (source == darkAverageRadio)
                darkMedian = false;
        else if (source == darkMedianRadio)
                darkMedian = true;
        else if (source == biasAverageRadio)
                biasMedian = false;
        else if (source == biasMedianRadio)
                biasMedian = true;
        else if (source == raTextField)
            {
            ra = sexToDec(raTextField);
            raTextField.setText(decToSex(ra, 3, false));
            }
        else if (source == decTextField)
            {
            dec = sexToDec(decTextField);
            decTextField.setText(decToSex(dec, 2, true));
            }
        else if (source == latTextField)
            {
            lat = sexToDec(latTextField);
            latTextField.setText(decToSex(lat, 2, false));
            }
        else if (source == lonTextField)
            {
            lon = sexToDec(lonTextField);
            lonTextField.setText(decToSex(lon, 2, false));
            }
		else if(source == dirButton)
			{
			if (running)
				{
				IJ.beep();
				return;
				}
			if (path.equals(""))
				path = IJ.getDirectory ("home");

            DirectoryChooser.setDefaultDirectory(path);
			DirectoryChooser od = new DirectoryChooser ("Select primary directory");

			if (od.getDirectory() != null)
				{
				path = od.getDirectory();
				dirText.setText (path);
				dirText.setHorizontalAlignment (JTextField.LEFT);
                images.clear();
                firstRun =true;
                restart = true;
                loadNewCals = true;
                foundImages = 0;
                ignoredImages = 0;
                updateStatusLabel();
                countValidFiles();
				}
			}
		else if(source == fileButton)
			{
			if (running)
				{
				IJ.beep();
				return;
				}
            filepath = path;
			if (filepath.equals(""))
				filepath = IJ.getDirectory ("home");
            OpenDialog.setDefaultDirectory(filepath);
            OpenDialog of = new OpenDialog("Select a file", "");

			if (of.getDirectory() != null)
				{
				path = of.getDirectory();
				dirText.setText (path);
				dirText.setHorizontalAlignment (JTextField.LEFT);
				}
			if (of.getFileName() != null)
				{
				filenamePattern = of.getFileName();
                filenamePattern = filenamePattern.replaceFirst("_.*\\.", "_*.");
				filenamePatternText.setText (filenamePattern);
				filenamePatternText.setHorizontalAlignment (JTextField.RIGHT);
                }
            if (of.getDirectory() != null || of.getFileName() != null)
                {
                images.clear();
                firstRun =true;
                restart = true;
                loadNewCals = true;
                foundImages = 0;
                ignoredImages = 0;
                updateStatusLabel();
                countValidFiles();
				}
			}
		else if(source == preMacroButton)
			{
			if (running)
				{
				IJ.beep();
				return;
				}
			if (preMacroPath.equals(""))
				preMacroPath = IJ.getDirectory ("home");
            int dirSize = preMacroPath.lastIndexOf(slash);
            OpenDialog.setDefaultDirectory(preMacroPath.substring(0, dirSize));
 			OpenDialog od = new OpenDialog ("Select pre-macro","");
			if (od.getDirectory() != null)
				{
				preMacroPath = od.getDirectory()+od.getFileName();
				preMacroText.setText (preMacroPath);
				preMacroText.setHorizontalAlignment (JTextField.RIGHT);
//                images.clear();
//                firstRun = true;
//                foundImages = 0;
//                ignoredImages = 0;
                updateStatusLabel();
				}
			}
		else if(source == postMacroButton)
			{
			if (running)
				{
				IJ.beep();
				return;
				}
			if (postMacroPath.equals(""))
				postMacroPath = IJ.getDirectory ("home");
            int dirSize = postMacroPath.lastIndexOf(slash);
            OpenDialog.setDefaultDirectory(postMacroPath.substring(0, dirSize));
 			OpenDialog od = new OpenDialog ("Select post-macro","");
			if (od.getDirectory() != null)
				{
				postMacroPath = od.getDirectory()+od.getFileName();
				postMacroText.setText (postMacroPath);
				postMacroText.setHorizontalAlignment (JTextField.RIGHT);
//                images.clear();
//                firstRun = true;
//                foundImages = 0;
//                ignoredImages = 0;
                updateStatusLabel();
				}
			}
        else if(source == stopButton)
			{
            stop();
			}
		else if(source == clearButton)
			{
			if (running)
				{
				IJ.beep();
				return;
				}
            else
                {
                images.clear();
                firstRun = true;
                restart = true;
                loadNewCals = true;
                foundImages = 0;
                ignoredImages = 0;
                startButton.setText("START");
                startButton.setForeground(Color.black);
                startButton.repaint();
                updateStatusLabel();
				}
			}
		else if(source == startButton)
			{
            if (running)
				{
				IJ.beep();
				return;
				}
            else
                {
                if (timer != null) timer.cancel();
                if (task != null) task.cancel();
                startButton.setText("RUNNING");
                startButton.setForeground(Color.red);
                startButton.repaint();

                length = 0;
                if (onlyNew) firstRun=true;
                running = true;
                requestStop = false;
                startTimer();
                }
			}
		}

public void propertyChange(PropertyChangeEvent e) {
    Object source = e.getSource();
//    if (source == raTextField)
//        {
//        ra = ((Number)raTextField.getValue()).doubleValue();
//        }
    }

protected MaskFormatter createFormatter(String s) {
    MaskFormatter formatter = null;
    try {
        formatter = new MaskFormatter(s);
        formatter.setOverwriteMode(true);
    } catch (java.text.ParseException exc) {
        System.err.println("formatter is bad: " + exc.getMessage());
        System.exit(-1);
    }
    return formatter;
    }


class thisDocumentListener implements DocumentListener
    {
    public void insertUpdate (DocumentEvent ev)
        {

        getTextFields();
        countValidFiles();
        }
    public void removeUpdate (DocumentEvent ev)
        {

        getTextFields();
        countValidFiles();
        }
    public void changedUpdate (DocumentEvent ev)
        {

        getTextFields();
        countValidFiles();
        }

    }
    void getTextFields()
        {
        path = dirText.getText();
        oldFilenamePattern = filenamePattern;
        filenamePattern = filenamePatternText.getText();
        if (updateExclude && !filenamePattern.equals(oldFilenamePattern))
            {
            if (filenamePattern.lastIndexOf('*') >= 0)
                {
                numPrefix = filenamePattern.substring(0, filenamePattern.lastIndexOf('*'));
                numPrefixField.setText(numPrefix);
                }
            else
                {
                numPrefix = filenamePattern;
                numPrefixField.setText(numPrefix);
                }
            }
        else
            {
            numPrefix = numPrefixField.getText();
            }
        preMacroPath = preMacroText.getText();
        postMacroPath = postMacroText.getText();
        biasDir = biasDirField.getText();
        biasBase = biasBaseField.getText();
        biasMaster = biasMasterField.getText();
        darkDir = darkDirField.getText();
        darkBase = darkBaseField.getText();
        darkMaster = darkMasterField.getText();
        flatDir = flatDirField.getText();
        flatBase = flatBaseField.getText();
        flatMaster = flatMasterField.getText();
        saveDir = saveDirField.getText();
        saveSuffix = saveSuffixField.getText();
        saveFormat = saveFormatField.getText();
        ra = sexToDec(raTextField);
        dec = sexToDec(decTextField);
        lat = sexToDec(latTextField);
        lon = sexToDec(lonTextField);
        loadNewCals = true;
        }

    double sexToDec(JTextField textField)
        {
        double value = 0.0;
        boolean isNegative = false;
        String text = textField.getText();
        String[] pieces = text.split("[0-9\\.]{1,}");
        if (pieces.length > 0 && pieces[0].endsWith("-")) isNegative = true;
        if (text.toUpperCase().contains("S")||text.toUpperCase().contains("W")) isNegative = !isNegative;
        pieces = text.replaceAll("[^0-9\\.]{1,}", " ").trim().split("[^0-9\\.]{1,}");
        try {if (pieces.length > 0) value += Double.parseDouble(pieces[0]);}
        catch (NumberFormatException e) { }
        try {if (pieces.length > 1) value += Double.parseDouble(pieces[1])/60.0;}
        catch (NumberFormatException e) { }
        try {if (pieces.length > 2) value += Double.parseDouble(pieces[2])/3600.0;}
        catch (NumberFormatException e) { }
        if (isNegative) value = -value;
        return value;
        }

	public String decToSex (double d, int fractionPlaces, Boolean showPlus)
		{
        DecimalFormat nf = new DecimalFormat();
        DecimalFormat nf2x =  new DecimalFormat();
		nf.setMinimumIntegerDigits(2);
        nf2x.setMinimumIntegerDigits(2);
        nf2x.setMinimumFractionDigits(0);
        nf2x.setMaximumFractionDigits(fractionPlaces);

		double dd = Math.abs(d);
        dd += 0.00000000000001;    //helps correct rounding issues when converting back to dms format

		int h = (int)dd;
		int m = (int)(60.0*(dd-(double)h));
		double s = 3600.0*(dd-(double)h-(double)m/60.0);

		String str = "";
		if (d < 0.0) str = "-";
        else if (showPlus) str = "+";
		str += ""+nf.format(h)+":"+nf.format(m)+":"+nf2x.format(s);
		return str;
		}


    boolean validateSelections()
        {
        path = path.trim();
        if (!path.endsWith(slash))
            {
            path += slash;
            dirText.setText(path);
            }

        scienceDir = new File (path);
        if (!scienceDir.isDirectory())
            {
            if (showLog) IJ.log("ERROR: Primary directory \""+path+"\"not found.");
            IJ.showMessage ("ERROR: Primary directory \""+path+"\"not found.");
            return false;
            }

        if (runPreMacro)
            {
            File preMacroFile = new File(preMacroText.getText().trim());
            if (!preMacroFile.isFile())
                {
                if (showLog) IJ.log("ERROR: Macro 1 \""+preMacroText.getText().trim()+"\" not found.");
                IJ.showMessage ("ERROR: Macro 1 \""+preMacroText.getText().trim()+"\" not found.");
                return false;
                }
            }

        if (runPostMacro)
            {
            File postMacroFile = new File(postMacroText.getText().trim());
            if (!postMacroFile.isFile())
                {
                if (showLog) IJ.log("ERROR: Macro 2 \""+postMacroText.getText().trim()+"\" not found.");
                IJ.showMessage ("ERROR: Macro 2 \""+postMacroText.getText().trim()+"\" not found.");
                return false;
                }
            }
        if (useBias || createBias)
            {
            biasPath = path + biasDir.trim();
            if (!biasPath.endsWith(slash)) biasPath += slash;
            biasDirectory = new File (biasPath);
            if (!biasDirectory.isDirectory())
                {
                if (showLog) IJ.log("ERROR: Bias directory \""+biasPath+"\" not found.");
                IJ.showMessage ("ERROR: Bias directory \""+biasPath+"\" not found.");
                return false;
                }
            }
        if (useBias && !createBias)
            {
            mbiasPath = biasPath + biasMaster.trim();
            biasMasterFile = new File (mbiasPath);
            if (!biasMasterFile.isFile())
                {
                if (showLog) IJ.log("ERROR: Master bias file \""+mbiasPath+"\" not found.");
                IJ.showMessage ("ERROR: Master bias file \""+mbiasPath+"\" not found.");
                return false;
                }
            }
        if (useDark || createDark)
            {
            darkPath = path + darkDir.trim();
            if (!darkPath.endsWith(slash)) darkPath += slash;
            darkDirectory = new File (darkPath);
            if (!darkDirectory.isDirectory())
                {
                if (showLog) IJ.log("ERROR: Dark directory \""+darkPath+"\" not found.");
                IJ.showMessage ("ERROR: Dark directory \""+darkPath+"\" not found.");
                return false;
                }
            }
        if (useDark && !createDark)
            {
            mdarkPath = darkPath + darkMaster.trim();
            darkMasterFile = new File (mdarkPath);
            if (!darkMasterFile.isFile())
                {
                if (showLog) IJ.log("ERROR: Master dark file \""+mdarkPath+"\" not found.");
                IJ.showMessage ("ERROR: Master dark file \""+mdarkPath+"\" not found.");
                return false;
                }
            }
        if (useFlat || createFlat)
            {
            flatPath = path + flatDir.trim();
            if (!flatPath.endsWith(slash)) flatPath += slash;
            flatDirectory = new File (flatPath);
            if (!flatDirectory.isDirectory())
                {
                if (showLog) IJ.log("ERROR: Flat directory \""+flatPath+"\" not found.");
                IJ.showMessage ("ERROR: Flat directory \""+flatPath+"\" not found.");
                return false;
                }
            }
        if (useFlat && !createFlat)
            {
            mflatPath = flatPath + flatMaster.trim();
            flatMasterFile = new File (mflatPath);
            if (!flatMasterFile.isFile())
                {
                if (showLog) IJ.log("ERROR: Master flat file \""+mflatPath+"\" not found.");
                IJ.showMessage ("ERROR: Master flat file \""+mflatPath+"\" not found.");
                return false;
                }
            }
        return true;
        }

    boolean createMasterBias()
        {
        ImageStack stack=null;
        ImagePlus imp=null, imp2=null;
        ImageProcessor mbiasIp;
        ImageProcessor[] slices = null;
        try
            {
            String bias;
            double min = 0, max = 0;
            int count = 0;
            biasPath = path + biasDir.trim();
            if (!biasPath.endsWith(slash)) biasPath += slash;
            biasDirectory = new File (biasPath);
            if (!biasDirectory.isDirectory())
                {
                if (showLog) IJ.log("ERROR: Bias directory \""+biasPath+"\" not found.");
                IJ.showMessage ("ERROR: Bias directory \""+biasPath+"\" not found.");
                return false;
                }
            File[] files = biasDirectory.listFiles ();
            String[] mbiasHeader = null;
            String[] filenames = new String[files.length];
            String[] validFilenames = new String[files.length];
            for (int i=0; i < files.length; i++)
                    {
                    filenames[i] = files[i].getName();
                    }

            if (sortNumerically) {filenames=sortFileList(filenames);}

            for (int i=0; i < files.length; i++)
                {
                bias = filenames[i];
                if (matchWildCard(biasBase, bias))
                    {
                    if (requestStop) {return false;}
                    count++;
                    if (useBeep) {IJ.beep();}
                    biasFilePath = biasPath+bias;
                    if (showLog) {IJ.log("Loading bias file \""+biasFilePath+"\" ("+count+" of "+validBiasFiles+")");}
                    imp = IJ.openImage(biasFilePath);
                    if (imp == null)
                        {
                        if (showLog) {IJ.log("ERROR: Unable to open image \""+biasFilePath+"\".");}
                        IJ.showMessage("ERROR: Unable to open image \""+biasFilePath+"\".");
                        return false;
                        }
                    else
                        {
                        if (count == 1)
                            {
                            mbiasHeader = FitsJ.getHeader(imp);
                            }
                        if (stack==null){
                            stack = imp.createEmptyStack();}

                        ImageProcessor ip = imp.getProcessor();
                        if (ip.getMin()<min) min = ip.getMin();
                        if (ip.getMax()>max) max = ip.getMax();
                        stack.addSlice(bias, ip);
                        FitsJ.putHeader(stack, FitsJ.getHeader(imp), count);

                        IJ.showStatus((stack.getSize()+1) + ": " + bias);
                        }
                    validFilenames[count-1] = filenames[i];
                    }
                }
            if (count == 0)
                {
                if (showLog) IJ.log("ERROR: No bias image matching \""+biasPath+biasBase.trim()+"\".");
                IJ.showMessage("ERROR: No bias image matching \""+biasPath+biasBase.trim()+"\".");
                return false;
                }
            if (stack!=null)
                {
                imp2 = WindowManager.getImage("Bias Stack");
                if (imp2 != null) imp2.close();
                imp2 = new ImagePlus("Bias Stack", stack);

                if (imp2.getBitDepth()==16 || imp.getBitDepth()==32)
                    imp2.getProcessor().setMinAndMax(min, max);
                if (showMasters) imp2.show();


    //            stack = imp2.getStack();
                int sliceCount = stack.getSize();
                slices = new ImageProcessor[sliceCount];
                for (int i=0; i<sliceCount; i++)
                    slices[i] = stack.getProcessor(i+1);
                mbiasIp = slices[0].duplicate();
                mbiasIp = mbiasIp.convertToFloat();

                int width = mbiasIp.getWidth();
                int height = mbiasIp.getHeight();
                int inc = Math.max(height/30, 1);
                if (requestStop) {return false;}
                if (biasMedian)
                    {
                    IJ.showStatus("Calculating median...");
                    mbiasHeader = FitsJ.addHistory("Median master bias created from "+count+" images", mbiasHeader);
                    if (showLog) {IJ.log("Calculating median master bias image \""+biasPath+biasMaster.trim()+"\".");}
                    float[] values = new float[sliceCount];
                    for (int y=0; y<height; y++)
                        {
                        if (y%inc==0) {IJ.showProgress(y, height-1);}
                        for (int x=0; x<width; x++)
                            {
                            for (int i=0; i<sliceCount; i++){
                                values[i] = slices[i].getPixelValue(x, y);}
                            java.util.Arrays.sort(values);
                            if ((sliceCount % 2) == 0)      // even
                                {mbiasIp.putPixelValue(x, y, (values[sliceCount / 2] + values[sliceCount / 2 - 1]) / 2.0);}
                            else                        //odd
                                {mbiasIp.putPixelValue(x, y, values[sliceCount / 2]);}
                            }

                        }
                    }
                else
                    {

                    IJ.showStatus("Calculating average...");
                    mbiasHeader = FitsJ.addHistory("Average master bias created from "+count+" images", mbiasHeader);
                    if (showLog) IJ.log("Calculating average master bias image \""+biasPath+biasMaster.trim()+"\".");
                    double value;
                    for (int y=0; y<height; y++)
                        {
                        if (y%inc==0) IJ.showProgress(y, height-1);
                        for (int x=0; x<width; x++)
                            {
                            value = 0.0;
                            for (int i=0; i<sliceCount; i++)
                                value += slices[i].getPixelValue(x, y);
                            mbiasIp.putPixelValue(x, y, (float)(value / (double)sliceCount));
                            }
                        }
                    }

                mbiasImp = WindowManager.getImage(biasMaster.trim());
                if (mbiasImp != null) mbiasImp.close();
                mbiasImp = new ImagePlus(biasMaster.trim(), mbiasIp);
                mbiasHeader = FitsJ.addHistory("on "+f.format(new java.util.Date())+" (YYYY-MM-DD hh:mm:ss UT)", mbiasHeader);
                for (int i=0; i<count; i++)
                    mbiasHeader = FitsJ.addHistory("Image "+(i+1)+" = "+biasPath+validFilenames[i], mbiasHeader);
    //            mbiasHeader = FitsJ.removeCards("DATE", mbiasHeader);
    //            mbiasHeader = FitsJ.setCard("DATE", f.format(new java.util.Date()), "file creation date (YYYY-MM-DD hh:mm:ss UT)", mbiasHeader);

                FitsJ.putHeader(mbiasImp, mbiasHeader);

                if (showMasters)
                    {
                    mbiasImp.show();
                    }
                if (requestStop) {return false;}
                saveProcessedFile(mbiasImp, biasPath, biasPath+biasMaster.trim(), "master bias", "");
                }
            else
                {
                IJ.showMessage("ERROR: Bias stack is empty.");
                if (showLog) IJ.log("ERROR: Bias stack is empty.");
                return false;
                }
            }
        catch (OutOfMemoryError e)
            {
            IJ.outOfMemory("Bias Stack");
            IJ.log ("ERROR: Out of memory");
            return false;
            }
        return true;
        }

    boolean createMasterDark()
        {
        ImageStack stack=null;
        ImagePlus imp=null, imp2=null;
        ImageProcessor mdarkIp;
        ImageProcessor mbiasIp = null;
        ImageProcessor[] slices = null;
        try
            {
            String dark;
            double min = 0, max = 0;
            int count = 0;
            darkPath = path + darkDir.trim();
            if (!darkPath.endsWith(slash)) darkPath += slash;
            darkDirectory = new File (darkPath);
            if (!darkDirectory.isDirectory())
                {
                if (showLog) IJ.log("ERROR: Dark directory \""+darkPath+"\" not found.");
                IJ.showMessage ("ERROR: Dark directory \""+darkPath+"\" not found.");
                return false;
                }
            File[] files = darkDirectory.listFiles ();
            String[] mdarkHeader = null;
            String[] filenames = new String[files.length];
            String[] validFilenames = new String[files.length];

            if (useBias && !loadMasterBiasFile())
                {
                return false;
                }
            if (useBias) mbiasIp = mbiasImp.getProcessor();
            for (int i=0; i < files.length; i++)
                    {
                    filenames[i] = files[i].getName();
                    }

            if (sortNumerically) filenames=sortFileList(filenames);

            for (int i=0; i < files.length; i++)
                {
                dark = filenames[i];
                if (matchWildCard(darkBase, dark))
                    {
                    if (requestStop) {return false;}
                    count++;
                    if (useBeep) IJ.beep();
                    darkFilePath = darkPath+dark;
                    if (showLog) IJ.log("Loading dark file \""+darkFilePath+"\" ("+count+" of "+validDarkFiles+")");
                    imp = IJ.openImage(darkFilePath);
                    if (imp == null)
                        {
                        if (showLog) IJ.log("ERROR: Unable to open image \""+darkFilePath+"\".");
                        IJ.showMessage("ERROR: Unable to open image \""+darkFilePath+"\".");
                        return false;
                        }
                    else
                        {
                        if (count == 1)
                            {
                            mdarkHeader = FitsJ.getHeader(imp);
                            }
                        if (stack==null)
                            stack = imp.createEmptyStack();
                        ImageProcessor ip = imp.getProcessor();
                        if (ip.getMin()<min) min = ip.getMin();
                        if (ip.getMax()>max) max = ip.getMax();
                        stack.addSlice(dark, ip);
                        FitsJ.putHeader(stack, FitsJ.getHeader(imp), count);

                        IJ.showStatus((stack.getSize()+1) + ": " + dark);

                        }
                    validFilenames[count-1] = filenames[i];
                    }
                }
            if (count == 0)
                {
                if (showLog) IJ.log("ERROR: No dark image matching \""+darkPath+darkBase.trim()+"\".");
                IJ.showMessage("ERROR: No dark image matching \""+darkPath+darkBase.trim()+"\".");
                return false;
                }
            if (stack!=null)
                {
                imp2 = WindowManager.getImage("Dark Stack");
                if (imp2 != null) imp2.close();
                imp2 = new ImagePlus("Dark Stack", stack);
                if (imp2.getBitDepth()==16 || imp.getBitDepth()==32)
                    imp2.getProcessor().setMinAndMax(min, max);
                if (showMasters) imp2.show();


    //            stack = imp2.getStack();
                int sliceCount = stack.getSize();
                slices = new ImageProcessor[sliceCount];
                for (int i=0; i<sliceCount; i++)
                    slices[i] = stack.getProcessor(i+1);
                mdarkIp = slices[0].duplicate();
                mdarkIp = mdarkIp.convertToFloat();

                int width = mdarkIp.getWidth();
                int height = mdarkIp.getHeight();
                int inc = Math.max(height/30, 1);
                if (requestStop) {return false;}
                if (darkMedian)
                    {
                    IJ.showStatus("Calculating median...");
                    mdarkHeader = FitsJ.addHistory("Median master dark created from "+count+" images", mdarkHeader);
                    if (showLog) IJ.log("Calculating median master dark image \""+darkPath+darkMaster.trim()+"\".");
                    double[] values = new double[sliceCount];
                    for (int y=0; y<height; y++)
                        {
                        if (y%inc==0) IJ.showProgress(y, height-1);
                        for (int x=0; x<width; x++)
                            {
                            for (int i=0; i<sliceCount; i++)
                                {
                                if (!useBias)
                                    {
                                    values[i] = (double)slices[i].getPixelValue(x, y);
                                    }
                                else
                                    {
                                    values[i] = (double)slices[i].getPixelValue(x, y) - (double)mbiasIp.getPixelValue(x, y);
                                    if (useNLC) values[i] = coeffA + values[i]*coeffB + values[i]*values[i]*coeffC;
                                    }
                                }
                            java.util.Arrays.sort(values);
                            if ((sliceCount % 2) == 0)      // even
                                mdarkIp.putPixelValue(x, y, (float)((values[sliceCount/2] + values[sliceCount/2 - 1]) / 2.0));
                            else                        //odd
                                mdarkIp.putPixelValue(x, y, (float)values[sliceCount/2]);
                            }

                        }
                    }
                else
                    {
                    IJ.showStatus("Calculating average...");
                    mdarkHeader = FitsJ.addHistory("Average master dark created from "+count+" images", mdarkHeader);
                    if (showLog) IJ.log("Calculating average master dark image \""+darkPath+darkMaster.trim()+"\".");
                    double value, pixel;
                    for (int y=0; y<height; y++)
                        {
                        if (y%inc==0) IJ.showProgress(y, height-1);
                        for (int x=0; x<width; x++)
                            {
                            value = 0.0;
                            for (int i=0; i<sliceCount; i++)
                                {
                                if (!useBias)
                                    value += (double)(slices[i].getPixelValue(x, y));
                                else if (useBias && !useNLC)
                                    value += (double)slices[i].getPixelValue(x, y) - (double)mbiasIp.getPixelValue(x, y);
                                else //useBias && useNLC
                                    {
                                    pixel = (double)slices[i].getPixelValue(x, y) - (double)mbiasIp.getPixelValue(x, y);
                                    value += coeffA + pixel*coeffB + pixel*pixel*coeffC;
                                    }
                                }
                            mdarkIp.putPixelValue(x, y, (float)(value / (double)sliceCount));
                            }
                        }
                    }

                mdarkImp = WindowManager.getImage(darkMaster.trim());
                if (mdarkImp != null) mdarkImp.close();
                mdarkImp = new ImagePlus(darkMaster.trim(), mdarkIp);
                mdarkHeader = FitsJ.addHistory("on "+f.format(new java.util.Date())+" (YYYY-MM-DD hh:mm:ss UT)", mdarkHeader);
                if (useBias)
                    {
                    mdarkHeader = FitsJ.addHistory("Bias corrected with "+biasMaster.trim(), mdarkHeader);
                    if (showLog) IJ.log("    Bias corrected with "+biasMaster.trim());
                    }
                if (useNLC & useBias)
                    {
                    mdarkHeader = FitsJ.addHistory("Non-linear corrected with coefficients:", mdarkHeader);
                    mdarkHeader = FitsJ.addHistory("a0 = "+coeffA, mdarkHeader);
                    mdarkHeader = FitsJ.addHistory("a1 = "+coeffB, mdarkHeader);
                    mdarkHeader = FitsJ.addHistory("a2 = "+coeffC, mdarkHeader);
                    if (showLog) IJ.log("    Non-linear corrected with coefficients: a0="+coeffA+" a1="+coeffB+" a2="+coeffC);
                    }
                for (int i=0; i<count; i++)
                    mdarkHeader = FitsJ.addHistory("Image "+(i+1)+" = "+darkPath+validFilenames[i], mdarkHeader);
    //            mdarkHeader = FitsJ.removeCards("DATE", mdarkHeader);
    //            mdarkHeader = FitsJ.setCard("DATE", f.format(new java.util.Date()), "file creation date (YYYY-MM-DD hh:mm:ss UT)", mdarkHeader);
                FitsJ.putHeader(mdarkImp, mdarkHeader);

                if (showMasters) mdarkImp.show();
                if (requestStop) {return false;}
                saveProcessedFile(mdarkImp, darkPath, darkPath+darkMaster.trim(), "master dark", "");
                }
            else
                {
                IJ.showMessage("ERROR: Dark stack is empty.");
                if (showLog) IJ.log("ERROR: Dark stack is empty.");
                return false;
                }
            }
        catch (OutOfMemoryError e)
            {
            IJ.outOfMemory("Dark Stack");
            IJ.log ("ERROR: Out of memory");
            return false;
            }
        return true;
        }

    boolean createMasterFlat()
        {
        ImageStack stack=null;
        ImagePlus imp=null, imp2=null;
        ImageProcessor[] slices = null;
        ImageProcessor mbiasIp = null;
        ImageProcessor mdarkIp = null;
        ImageProcessor mflatIp = null;
        try
            {
            String flat;
            double min = 0, max = 0;
            int count = 0;
            flatPath = path + flatDir.trim();
            if (!flatPath.endsWith(slash)) flatPath += slash;
            flatDirectory = new File (flatPath);
            if (!flatDirectory.isDirectory())
                {
                if (showLog) IJ.log("ERROR: Flat directory \""+flatPath+"\" not found.");
                IJ.showMessage ("ERROR: Flat directory \""+flatPath+"\" not found.");
                return false;
                }
            File[] files = flatDirectory.listFiles ();
            String[] mflatHeader = null;
            String[] fHeader = null;
            String[] dHeader = null;
            String[] filenames = new String[files.length];
            String[] validFilenames = new String[files.length];
            double[] expTimeFactor = new double[files.length];
            double dExpTime = 1;
            double fExpTime = 1;

            if (useBias && !loadMasterBiasFile())
                {
                return false;
                }
            if (useBias) mbiasIp = mbiasImp.getProcessor();
            if (requestStop) {return false;}
            if (useDark && !loadMasterDarkFile())
                {
                return false;
                }
            if (useDark) mdarkIp = mdarkImp.getProcessor();
            if (requestStop) {return false;}
            if (scaleExpTime && useDark)
                {
                dHeader = FitsJ.getHeader(mdarkImp);
                if (dHeader == null)
                    {
                    if (showLog) IJ.log("ERROR: Cannot extract FITS header from master dark image \""+mdarkPath+"\".");
                    IJ.showMessage("ERROR: Cannot extract FITS header from master dark image \""+mdarkPath+"\".");
                    return false;
                    }
                dExpTime = (double)FitsJ.getExposureTime(dHeader);
                if (Double.isNaN(dExpTime))
                    {
                    if (showLog) IJ.log("ERROR: Cannot extract exposure time from \""+mdarkPath+"\" master dark header!");
                    IJ.showMessage("ERROR: Cannot extract exposure time from \""+mdarkPath+"\" master dark header!");
                    return false;
                    }
                }

            for (int i=0; i < files.length; i++)
                    {
                    filenames[i] = files[i].getName();
                    }

            if (sortNumerically) filenames=sortFileList(filenames);

            for (int i=0; i < files.length; i++)
                {
                flat = filenames[i];
                expTimeFactor[i] = 1;
                if (matchWildCard(flatBase, flat))
                    {
                    if (requestStop) {return false;}
                    count++;
                    if (useBeep) IJ.beep();
                    flatFilePath = flatPath+flat;
                    if (showLog) IJ.log("Loading flat file \""+flatFilePath+"\" ("+count+" of "+validFlatFiles+")");
                    imp = IJ.openImage(flatFilePath);
                    if (imp == null)
                        {
                        if (showLog) IJ.log("ERROR: Unable to open image \""+flatFilePath+"\".");
                        IJ.showMessage("ERROR: Unable to open image \""+flatFilePath+"\".");
                        return false;
                        }
                    else
                        {

                        if (count == 1)
                            {
                            mflatHeader = FitsJ.getHeader(imp);
                            }

                        if (scaleExpTime && useDark)
                            {
                            fHeader = FitsJ.getHeader(imp);
                            if (fHeader == null)
                                {
                                if (showLog) IJ.log("ERROR: Cannot extract FITS header from flat image \""+flatFilePath+"\".");
                                IJ.showMessage("ERROR: Cannot extract FITS header from flat image \""+flatFilePath+"\".");
                                return false;
                                }
                            fExpTime = (double)FitsJ.getExposureTime(fHeader);
                            if (Double.isNaN(fExpTime))
                                {
                                if (showLog) IJ.log("ERROR: Cannot extract exposure time from flat image header!");
                                IJ.showMessage("ERROR: Cannot extract exposure time from flat image header!");
                                return false;
                                }
                            expTimeFactor[count-1] = fExpTime/dExpTime;
                            }

                        if (stack==null)
                            stack = imp.createEmptyStack();

                        ImageProcessor ip = imp.getProcessor();
                        ip = ip.convertToFloat();
                        if (ip.getMin()<min) min = ip.getMin();
                        if (ip.getMax()>max) max = ip.getMax();
                        stack.addSlice(flat, ip);
                        FitsJ.putHeader(stack, FitsJ.getHeader(imp), count);

                        IJ.showStatus((stack.getSize()+1) + ": " + flat);

                        }
                    validFilenames[count-1] = filenames[i];
                    }
                }
            if (count == 0)
                {
                if (showLog) IJ.log("ERROR: No flat image matching \""+flatPath+flatBase.trim()+"\".");
                IJ.showMessage("ERROR: No flat image matching \""+flatPath+flatBase.trim()+"\".");
                return false;
                }
            if (stack!=null)
                {
                imp2 = WindowManager.getImage("Flat Stack");
                if (imp2 != null) imp2.close();
                imp2 = new ImagePlus("Flat Stack", stack);
                if (imp2.getBitDepth()==16 || imp.getBitDepth()==32)
                    imp2.getProcessor().setMinAndMax(min, max);
                if (showMasters) imp2.show();


    //            stack = imp2.getStack();
                int sliceCount = stack.getSize();
                slices = new ImageProcessor[sliceCount];
                for (int i=0; i<sliceCount; i++)
                    slices[i] = stack.getProcessor(i+1);
                mflatIp = slices[0].duplicate();
                mflatIp = mflatIp.convertToFloat();

                int width = mflatIp.getWidth();
                int height = mflatIp.getHeight();
                int inc = Math.max(height/30, 1);

                if (useBias || useDark)
                    {
                    String operations = "";
                    if (useBias) operations = "Bias";
                    if (useBias && useNLC)
                        {
                        if (useDark) operations += ", Nonlinearity,";
                        else operations += " and Nonlinearity ";
                        }
                    if (useBias && useDark) operations += " and ";
                    if (useBias && !useNLC && !useDark) operations += " ";
                    if (useDark) operations += "Dark ";
                    operations += "Correction on:";
                    IJ.showStatus(""+operations);
                    if (showLog) IJ.log("Performing "+ operations);
                    double value;
                    for (int i=0; i<sliceCount; i++)
                        {
                        if (showLog) IJ.log("    "+validFilenames[i]+" using dark exposure scaling factor = "
                                +uptoFourPlaces.format(expTimeFactor[i]));
                        IJ.showProgress(i, sliceCount-1);
                        for (int y=0; y<height; y++)
                            {
                            for (int x=0; x<width; x++)
                                {
                                value = (double)(slices[i].getPixelValue(x, y));
                                if (useBias) value -= (double)mbiasIp.getPixelValue(x, y);
                                if (useBias && useNLC) value = coeffA + value*coeffB + value*value*coeffC;
                                if (useDark) value -= (double)mdarkIp.getPixelValue(x, y)*expTimeFactor[i];
                                slices[i].putPixelValue(x, y, value);
                                }

                            }
                        }
                    }
                if (showMasters) imp2.updateAndDraw();
                if (requestStop) {return false;}
                if (useGradientRemoval)
                    {
                    IJ.showStatus("Removing gradients ...");
                    for (int i=0; i<sliceCount; i++)
                        {
                        if (showLog) IJ.log("Removing gradient from calibrated flat "+validFilenames[i]+".");
                        IJ.showProgress(i, sliceCount-1);
                        slices[i] = removeGradient(slices[i]);
                        }
                    }
                if (showMasters) imp2.updateAndDraw();
                if (requestStop) {return false;}
                IJ.showStatus("Normalizing images ...");
                for (int i=0; i<sliceCount; i++)
                    {
                    IJ.showProgress(i, sliceCount-1);
                    slices[i] = normalizeImage(slices[i], validFilenames[i]);
                    if (slices[i] == null)
                        {
                        if (showLog) IJ.log("ERROR: image mean value = 0 for "+validFilenames[i]+". Divide Error. Normalize Failed.");
                        IJ.showMessage("ERROR: image mean value = 0 for "+validFilenames[i]+". Divide Error. Normalize Failed.");
                        return false;
                        }
                    }
                if (showMasters) imp2.updateAndDraw();
                if (requestStop) {return false;}
                if (flatMedian)
                    {
                    IJ.showStatus("Calculating median...");
                    mflatHeader = FitsJ.addHistory("Median master flat created from "+count+" images", mflatHeader);
                    if (showLog) IJ.log("Calculating median master flat image \""+flatPath+flatMaster.trim()+"\".");
                    double[] values = new double[sliceCount];
                    for (int y=0; y<height; y++)
                        {
                        if (y%inc==0) IJ.showProgress(y, height-1);
                        for (int x=0; x<width; x++)
                            {
                            for (int i=0; i<sliceCount; i++)
                                values[i] = (double)slices[i].getPixelValue(x, y);
                            java.util.Arrays.sort(values);
                            if ((sliceCount % 2) == 0)      // even
                                mflatIp.putPixelValue(x, y, (float)((values[sliceCount/2] + values[sliceCount/2 - 1]) / 2.0));
                            else                        //odd
                                mflatIp.putPixelValue(x, y, (float)values[sliceCount/2]);
                            }

                        }
                    }
                else
                    {
                    IJ.showStatus("Calculating average...");
                    mflatHeader = FitsJ.addHistory("Average master flat created from "+count+" images", mflatHeader);
                    if (showLog) IJ.log("Calculating average master flat image \""+flatPath+flatMaster.trim()+"\".");
                    double value;
                    for (int y=0; y<height; y++)
                        {
                        if (y%inc==0) IJ.showProgress(y, height-1);
                        for (int x=0; x<width; x++)
                            {
                            value = 0.0;
                            for (int i=0; i<sliceCount; i++)
                                value += (double)(slices[i].getPixelValue(x, y));
                            mflatIp.putPixelValue(x, y, (float)(value / (double)sliceCount));
                            }
                        }
                    }

                mflatImp = WindowManager.getImage(flatMaster.trim());
                if (mflatImp != null) mflatImp.close();
                mflatImp = new ImagePlus(flatMaster.trim(), mflatIp);
                mflatHeader = FitsJ.addHistory("on "+f.format(new java.util.Date())+" (YYYY-MM-DD hh:mm:ss UT)", mflatHeader);
                if (useBias)
                    {
                    mflatHeader = FitsJ.addHistory("Bias corrected with "+biasMaster.trim(), mflatHeader);
                    if (showLog) IJ.log("    Bias corrected with "+biasMaster.trim());
                    }
                if (useNLC & useBias)
                    {
                    mflatHeader = FitsJ.addHistory("Non-linear corrected with coefficients:", mflatHeader);
                    mflatHeader = FitsJ.addHistory("a0 = "+coeffA, mflatHeader);
                    mflatHeader = FitsJ.addHistory("a1 = "+coeffB, mflatHeader);
                    mflatHeader = FitsJ.addHistory("a2 = "+coeffC, mflatHeader);
                    if (showLog) IJ.log("    Non-linear corrected with coefficients: a0="+coeffA+" a1="+coeffB+" a2="+coeffC);
                    }
                if (useDark)
                    {
                    if (scaleExpTime)
                        {
                        mflatHeader = FitsJ.addHistory("Dark corrected with "+darkMaster.trim()+" with exposure time scaling", mflatHeader);
                        if (showLog) IJ.log("    Dark corrected with "+darkMaster.trim()+" with exposure time scaling");
                        }
                    else
                        {
                        mflatHeader = FitsJ.addHistory("Dark corrected with "+darkMaster.trim(), mflatHeader);
                        if (showLog) IJ.log("    Dark corrected with "+darkMaster.trim());
                        }
                    }
                if (useGradientRemoval)
                    {
                    mflatHeader = FitsJ.addHistory("Gradient removed from calibrated flats", mflatHeader);
                    if (showLog) IJ.log("    Gradient removed from calibrated flats");
                    }

                mflatHeader = FitsJ.addHistory("Normalized calibrated flats", mflatHeader);
                if (showLog) IJ.log("    Normalized calibrated flats");

                for (int i=0; i<count; i++)
                    mflatHeader = FitsJ.addHistory("Image "+(i+1)+" = "+flatPath+validFilenames[i], mflatHeader);
    //            mflatHeader = FitsJ.removeCards("DATE", mflatHeader);
    //            mflatHeader = FitsJ.setCard("DATE", f.format(new java.util.Date()), "file creation date (YYYY-MM-DD hh:mm:ss UT)", mflatHeader);
                FitsJ.putHeader(mflatImp, mflatHeader);

                if (showMasters) mflatImp.show();
                if (requestStop) {return false;}
                saveProcessedFile(mflatImp, flatPath, flatPath+flatMaster.trim(), "master flat", "");
                }
            else
                {
                IJ.showMessage("ERROR: Flat stack is empty.");
                if (showLog) IJ.log("ERROR: Flat stack is empty.");
                 return false;
                }
            }
        catch (OutOfMemoryError e)
            {
            IJ.outOfMemory("Flat Stack");
            IJ.log ("ERROR: Out of memory");
            return false;
            }
        return true;
        }



    boolean loadMasterBiasFile()
            {
            if (requestStop) return false;
            mbiasImp = WindowManager.getImage(biasMaster.trim());
            if (mbiasImp != null) mbiasImp.close();
            biasPath = path + biasDir.trim();
            if (!biasPath.endsWith(slash)) biasPath += slash;
            mbiasPath = biasPath + biasMaster.trim();
            if (showLog) IJ.log("Loading master bias file \""+mbiasPath+"\"");
            mbiasImp = IJ.openImage(mbiasPath);
            if (mbiasImp == null)
                {
                if (showLog) IJ.log("ERROR: Master bias file \""+mbiasPath+"\" open failed.");
                IJ.showMessage ("ERROR: Master bias file \""+mbiasPath+"\" open failed.");
                return false;
                }
            if (showMasters) mbiasImp.show();
            return true;
            }


    boolean loadMasterDarkFile()
            {
            if (requestStop) return false;
            mdarkImp = WindowManager.getImage(darkMaster.trim());
            if (mdarkImp != null) mdarkImp.close();
            darkPath = path + darkDir.trim();
            if (!darkPath.endsWith(slash)) darkPath += slash;
            mdarkPath = darkPath + darkMaster.trim();
            if (showLog) IJ.log("Loading master dark file \""+mdarkPath+"\"");
            mdarkImp = IJ.openImage(mdarkPath);
            if (mdarkImp == null)
                {
                if (showLog) IJ.log("ERROR: Master dark file \""+mdarkPath+"\" open failed.");
                IJ.showMessage ("ERROR: Master dark file \""+mdarkPath+"\" open failed.");
                return false;
                }
            if (showMasters) mdarkImp.show();
            return true;
            }


    boolean loadMasterFlatFile()
            {
            if (requestStop) return false;
            mflatImp = WindowManager.getImage(flatMaster.trim());
            if (mflatImp != null) mflatImp.close();
            flatPath = path + flatDir.trim();
            if (!flatPath.endsWith(slash)) flatPath += slash;
            mflatPath = flatPath + flatMaster.trim();
            if (showLog) IJ.log("Loading master flat file \""+mflatPath+"\"");
            mflatImp = IJ.openImage(mflatPath);
            if (mflatImp == null)
                {
                if (showLog) IJ.log("ERROR: Master flat file \""+mflatPath+"\" open failed.");
                IJ.showMessage ("ERROR: Master flat file \""+mflatPath+"\" open failed.");
                return false;
                }
            if (showMasters) mflatImp.show();

            return true;
            }


    boolean processData()
        {
        double pixel = 0.0;
        double expTimeFactor = 1.0;
        ImageProcessor scienceIp = scienceImp.getProcessor();
        scienceIp = scienceIp.convertToFloat();
        float[] sciencePix = (float[])scienceIp.getPixels();

        ImageProcessor mbiasIp = null;
        float[] mbiasPix = null;
        if (useBias)
            {
            mbiasIp = mbiasImp.getProcessor();
            mbiasIp = mbiasIp.convertToFloat();
            mbiasPix = (float[])mbiasIp.getPixels();
            }

        ImageProcessor mdarkIp = null;
        float[] mdarkPix = null;
        if (useDark)
            {
            mdarkIp = mdarkImp.getProcessor();
            mdarkIp = mdarkIp.convertToFloat();
            mdarkPix = (float[])mdarkIp.getPixels();
            }

        ImageProcessor mflatIp = null;
        float[] mflatPix = null;
        if (useFlat)
            {
            mflatIp = mflatImp.getProcessor();
            mflatIp = mflatIp.convertToFloat();
            mflatPix = (float[])mflatIp.getPixels();
            }

        scienceHeight = scienceIp.getHeight();
        scienceWidth = scienceIp.getWidth();
        length = scienceWidth * scienceHeight;
        if (useBias && mbiasIp.getHeight() != scienceHeight)
                {
                if (showLog) IJ.log("ERROR: Master bias height is not equal to \""+sciencePath+"\" height.");
                IJ.showMessage("ERROR: Master bias height is not equal to \""+sciencePath+"\" height.");
                return false;
                }
        if (useBias && mbiasIp.getWidth() != scienceWidth)
                {
                if (showLog) IJ.log("ERROR: Master bias width is not equal to \""+sciencePath+"\" width.");
                IJ.showMessage("ERROR: Master bias width is not equal to \""+sciencePath+"\" width.");
                return false;
                }
        if (useDark && mdarkIp.getHeight() != scienceHeight)
                {
                if (showLog) IJ.log("ERROR: Master dark height is not equal to \""+sciencePath+"\" height.");
                IJ.showMessage("ERROR: Master dark height is not equal to \""+sciencePath+"\" height.");
                return false;
                }
        if (useDark && mdarkIp.getWidth() != scienceWidth)
                {
                if (showLog) IJ.log("ERROR: Master dark width is not equal to \""+sciencePath+"\" width.");
                IJ.showMessage("ERROR: Master dark width is not equal to \""+sciencePath+"\" width.");
                return false;
                }
        if (useFlat && mflatIp.getHeight() != scienceHeight)
                {
                if (showLog) IJ.log("ERROR: Master flat height is not equal to \""+sciencePath+"\" height.");
                IJ.showMessage("ERROR: Master flat height is not equal to \""+sciencePath+"\" height.");
                return false;
                }
        if (useFlat && mflatIp.getWidth() != scienceWidth)
                {
                if (showLog) IJ.log("ERROR: Master flat width is not equal to \""+sciencePath+"\" width.");
                IJ.showMessage("ERROR: Master flat width is not equal to \""+sciencePath+"\" width.");
                return false;
                }

        scienceHeader = FitsJ.getHeader(scienceImp);
        if (scienceHeader == null)
            {
            if (showLog) IJ.log("ERROR: Cannot extract FITS header from science image \""+sciencePath+"\".");
            IJ.showMessage("ERROR: Cannot extract FITS header from science image \""+sciencePath+"\".");
            return false;
            }

        if (calcAirmass)
            {
            String JDType = "Mid Exp JD";
            String AMType = "mid-exposure";
            double jd = FitsJ.getMeanJD (scienceHeader);
            if (Double.isNaN(jd))
                {
                JDType = "Start Exp JD";
                AMType = "start of exposure";
                jd = FitsJ.getJD (scienceHeader);
                if (Double.isNaN(jd))
                    {
                    if (showLog) IJ.log("Couldn't extract JD from Fits header. Airmass calculation aborted.");
                    IJ.showMessage("Couldn't extract JD from Fits header. Airmass calculation aborted.");
                    }
                }
            if (!Double.isNaN(jd))
                {
                double hjdc = SkyCalc.getHJDCorrection(jd, ra, dec);
                double hjd = jd + hjdc;
                double altitude = SkyCalc.getAltitude(jd, ra, dec, lat, lon);
                double airmass = SkyCalc.getAirmass(jd, ra, dec, lat, lon);
                if (showLog)
                    {
                    IJ.log("    " + JDType + "=" + sixPlaces.format(jd)+
                       "  HJD="+sixPlaces.format(hjd)+
                       "  (correction = "+fourPlaces.format(hjdc*24*60)+" minutes)");
                    IJ.log("    Altitude="+twoPlaces.format(altitude)+
                       "  Airmass="+fourPlaces.format(airmass));
                    }
                if (saveAirmass)
                    {
                    scienceHeader = FitsJ.setCard("JD_SOBS", FitsJ.getJD (scienceHeader), "Julian Date at start of exposure (IJ)", scienceHeader);
                    if (JDType.equals("Mid Exp JD"))
                        {
                        scienceHeader = FitsJ.setCard("JD_MOBS", jd, "Julian Date at mid-exposure (IJ)", scienceHeader);
                        scienceHeader = FitsJ.setCard("HJD_MOBS", hjd, "Heliocentric JD at mid-exposure (IJ)", scienceHeader);
                        }
                    else
                        {
                        scienceHeader = FitsJ.setCard("HJD_SOBS", hjd, "Heliocentric JD at start of exposure (IJ)", scienceHeader);
                        }
                    scienceHeader = FitsJ.setCard("RA_OBJ", ra,
                                                  "right ascension of target (hours) (IJ)", scienceHeader);
                    scienceHeader = FitsJ.setCard("RA_OBJ", ra,
                                                  "right ascension of target (hours) (IJ)", scienceHeader);
                    scienceHeader = FitsJ.setCard("DEC_OBJ", dec,
                                                  "declination of target (degrees) (IJ)", scienceHeader);
                    scienceHeader = FitsJ.setCard("LATITUDE", lat,
                                                  "geographic latitude of observation (IJ)", scienceHeader);
                    scienceHeader = FitsJ.setCard("LONGITUD", lon,
                                                  "geographic longitude of observation (IJ)", scienceHeader);
                    scienceHeader = FitsJ.setCard("ALT_OBJ", altitude,
                                                  "altitude of target object at "+AMType+" (IJ)", scienceHeader);
                    scienceHeader = FitsJ.setCard("AIRMASS", airmass,
                                                  "airmass at "+AMType+" (IJ)", scienceHeader);
                    }
                }
            }

        if (scaleExpTime && useDark)
            {
            scienceExpTime = (double)FitsJ.getExposureTime(scienceHeader);
            if (Double.isNaN(scienceExpTime))
                {
                if (showLog) IJ.log("ERROR: Cannot extract exposure time from science image header!");
                IJ.showMessage("ERROR: Cannot extract exposure time from science image header!");
                return false;
                }
            darkHeader = FitsJ.getHeader(mdarkImp);
            if (darkHeader == null)
                {
                if (showLog) IJ.log("ERROR: Cannot extract FITS header from master dark image \""+mdarkPath+"\".");
                IJ.showMessage("ERROR: Cannot extract FITS header from master dark image \""+mdarkPath+"\".");
                return false;
                }
            darkExpTime = (double)FitsJ.getExposureTime(darkHeader);
            if (Double.isNaN(darkExpTime))
                {
                if (showLog) IJ.log("ERROR: Cannot extract exposure time from \""+mdarkPath+"\" master dark header!");
                IJ.showMessage("ERROR: Cannot extract exposure time from \""+mdarkPath+"\" master dark header!");
                return false;
                }
            expTimeFactor = scienceExpTime/darkExpTime;
            }

        for (int i = 0; i < length; i++)
            {
            pixel = (double)sciencePix[i];
            if (useBias)
                {
                pixel -= (double)mbiasPix[i];
                if (useNLC) pixel = coeffA + pixel*coeffB + pixel*pixel*coeffC;
                }
            if (useDark) pixel -= (double)mdarkPix[i]*expTimeFactor;
            if (useFlat)
                if (mflatPix[i] == 0.0f)
                    {
                    pixel = 65535.0;
                    if (useNLC && useBias)
                        pixel = coeffA + pixel*coeffB + pixel*pixel*coeffC;
                    }
                else
                    pixel /= (double)mflatPix[i];
            sciencePix[i]=(float)pixel;
            }

        if (!saveFloatingPoint)
            {
            scienceIp = scienceIp.convertToShort(false);
            }

        savePath = path + saveDir.trim();
        if (!savePath.endsWith(slash)) savePath += slash;
        saveDirPath = savePath;
        saveFileName = "Processed_"+s;

        if (saveProcessedData && (saveSuffix.trim().length()!=0 || saveDir.trim().length() != 0))
            {
            int dotIndex = s.lastIndexOf(".");
            if (dotIndex != -1) //a filetype exists
                {
                String filetype = s.substring(dotIndex);
                saveFileName = s.replace(filetype, ""+saveSuffix.trim()+filetype);
                }
            else
                saveFileName = s.concat(""+saveSuffix.trim()+".fits");
            }
        savePath += saveFileName;
        scienceHeader = FitsJ.addHistory("Previous Filename = "+s, scienceHeader);
        s = saveFileName;
        scienceImp.setProcessor(s, scienceIp);

        if (useBias)
            {
            scienceHeader = FitsJ.addHistory("Bias corrected with "+biasMaster.trim(), scienceHeader);
            if (showLog) IJ.log("    Bias corrected with "+biasMaster.trim());
            }
        if (useNLC & useBias)
            {
            scienceHeader = FitsJ.addHistory("Non-linear corrected with coefficients:", scienceHeader);
            scienceHeader = FitsJ.addHistory("a0 = "+coeffA, scienceHeader);
            scienceHeader = FitsJ.addHistory("a1 = "+coeffB, scienceHeader);
            scienceHeader = FitsJ.addHistory("a2 = "+coeffC, scienceHeader);
            if (showLog) IJ.log("    Non-linear corrected with coefficients: a0="+coeffA+" a1="+coeffB+" a2="+coeffC);
            }
        if (useDark)
            {
            scienceHeader = FitsJ.addHistory("Dark corrected with "+darkMaster.trim(), scienceHeader);
            if (showLog && !scaleExpTime) IJ.log("    Dark corrected with "+darkMaster.trim());
            }
        if (useDark && scaleExpTime)
            {
            scienceHeader = FitsJ.addHistory("and exposure time scaling factor = "+expTimeFactor, scienceHeader);
            if (showLog) IJ.log("    Dark corrected with "+darkMaster.trim()+" and exposure time scaling factor "+scienceExpTime+"/"+darkExpTime+"="+expTimeFactor);
            }
        if (useFlat)
            {
            scienceHeader = FitsJ.addHistory("Flat corrected with " + flatMaster.trim(), scienceHeader);
            if (showLog) IJ.log("    Flat corrected with " + flatMaster.trim());
            }


        FitsJ.putHeader(scienceImp, scienceHeader);

        IJ.wait(100);      //attempt to work around crash problem
        return true;
        }

    boolean saveProcessedFile(ImagePlus impLocal, String dirPath, String filePath, String name, String format)
        {
        File saveDirectory;
        if (impLocal != null)
            {
            saveDirectory = new File (dirPath);
            if (saveDirectory.isFile())
                {
                if (showLog) IJ.log("ERROR: Save directory \""+dirPath+"\" is a file, not a directory.");
                IJ.showMessage ("ERROR: Save directory \""+dirPath+"\" is a file, not a directory.");
                return false;
                }
            if (!saveDirectory.isDirectory())
                {
                if (!saveDirectory.mkdir())
                    {
                    if (showLog) IJ.log("ERROR: Could not create save-to directory \""+dirPath+"\".");
                    IJ.showMessage ("ERROR: Could not create save-to directory \""+dirPath+"\".");
                    return false;
                    }

                }
            if (!format.trim().equals("") && !format.trim().equals(null))
                {
                filePath = updateExtension(filePath, format);
                }

            if (filePath.endsWith(".fit") || filePath.endsWith(".FIT") || filePath.endsWith(".FITS") ||
                    filePath.endsWith(".fts") || filePath.endsWith(".FTS"))
                {
                IJ.runPlugIn(impLocal, "ij.plugin.FITS_Writer", filePath);
                }
            else
                {
                IJ.save(impLocal, filePath);
                }
            if (showLog) IJ.log("    Saved "+name+" file \""+filePath+"\"");
            }
        else
            {
            if (showLog) IJ.log("ERROR: "+name+" image is invalid prior to save.");
            IJ.showMessage("ERROR: "+name+" image is invalid prior to save.");
            return false;
            }
        return true;
        }

	static String updateExtension(String path, String extension)
        {
		if (path==null) return null;
		int dotIndex = path.lastIndexOf(".");
		int separatorIndex = path.lastIndexOf(File.separator);
		if (dotIndex>=0 && dotIndex>separatorIndex && (path.length()-dotIndex)<=5) {
			if (dotIndex+1<path.length() && Character.isDigit(path.charAt(dotIndex+1)))
				path += extension;
			else
				path = path.substring(0, dotIndex) + extension;
		} else
			path += extension;
		return path;
        }

    void updateStatusLabel()
        {
        statusLabel.setText ("processed "+foundImages+", ignored "+ignoredImages+"  ");
        statusLabel.repaint();
        }

    void stop()
        {
        unlock();
        requestStop = true;
        if (timer != null) timer.cancel();
        if (task != null) task.cancel();
        if (running)
            {
            startButton.setText("CONTINUE");
            startButton.setForeground(Color.orange);
            startButton.repaint();
            }
        running = false;
        }

    void reset()
        {
        String message;
        unlock();
        if (requestStop) message = "STOPPED";
        else message = "FINISHED";
        requestStop = true;
        if (timer != null) timer.cancel();
        if (task != null) task.cancel();
        running = false;

        images.clear();
        firstRun = true;
        restart = true;
        loadNewCals = true;
        foundImages = 0;
        ignoredImages = 0;
        startButton.setText("START");
        startButton.setForeground(Color.black);
        startButton.repaint();
        updateStatusLabel();
        if (showLog)
            {
            IJ.log(" ");
            IJ.log("************"+message+"************");
            IJ.log(" ");
            }
        }



    void countValidFiles()
        {
        validFiles = 0;
        path = path.trim();
        if (!path.endsWith(slash))
            path += slash;

        File sDir = new File (path);
        if (sDir.isDirectory())
            {
            File[] files = sDir.listFiles();
            if (files.length > 0)
                {
                String[] filenames = new String[files.length];
                for (int i=0; i < files.length; i++)
                        {
                        filenames[i] = files[i].getName();
                        }

                for (int i=0; i < files.length; i++)
                    {
                    String ss = filenames[i];
                    if (matchWildCard(filenamePattern, ss) && stringLongVal(ss) >= minFileNumber && stringLongVal(ss) <= maxFileNumber)
                        {
                        validFiles++;
                        }
                    }
                }
            }
        validFilesLabel.setText ("   "+validFiles+" Matching Files");
        validFilesLabel.repaint();


        String bPath = path + biasDir.trim();
        if (!bPath.endsWith(slash)) bPath += slash;
        validBiasFiles = cntFiles(bPath, biasBase.trim());
        validBiasFilesLabel.setText (""+validBiasFiles);
        validBiasFilesLabel.repaint();
        validMasterBiasFiles = cntFiles(bPath, biasMaster);
        validMasterBiasFilesLabel.setText (""+validMasterBiasFiles);
        validMasterBiasFilesLabel.repaint();

        String dPath = path + darkDir.trim();
        if (!dPath.endsWith(slash)) dPath += slash;
        validDarkFiles = cntFiles(dPath, darkBase.trim());
        validDarkFilesLabel.setText (""+validDarkFiles);
        validDarkFilesLabel.repaint();
        validMasterDarkFiles = cntFiles(dPath, darkMaster);
        validMasterDarkFilesLabel.setText (""+validMasterDarkFiles);
        validMasterDarkFilesLabel.repaint();

        String fPath = path + flatDir.trim();
        if (!fPath.endsWith(slash)) fPath += slash;
        validFlatFiles = cntFiles(fPath, flatBase.trim());
        validFlatFilesLabel.setText (""+validFlatFiles);
        validFlatFilesLabel.repaint();
        validMasterFlatFiles = cntFiles(fPath, flatMaster);
        validMasterFlatFilesLabel.setText (""+validMasterFlatFiles);
        validMasterFlatFilesLabel.repaint();


        return;
        }
    
    int cntFiles(String calPath, String calBase)
        {
        int cnt = 0;
        File calDir = new File (calPath);
        if (calDir.isDirectory())
            {
            File[] files = calDir.listFiles();
            if (files.length > 0)
                {
                String[] filenames = new String[files.length];
                for (int i=0; i < files.length; i++)
                        {
                        filenames[i] = files[i].getName();
                        }

                for (int i=0; i < files.length; i++)
                    {
                    String ss = filenames[i];
                    if (matchWildCard(calBase, ss))
                        {
                        cnt++;
                        }
                    }
                }
            }
        return cnt;
        }

//    boolean matchWildCard(String wildCard, String filename)
//        {
//        return filename.startsWith(wildCard.trim()) && filename.endsWith(".fits");
//        }


        void getPrefs()
        {
		path = Prefs.get ("watcher.path",path);
		filenamePattern = Prefs.get ("watcher.filenamePattern",filenamePattern);
		preMacroPath = Prefs.get ("watcher.preMacroPath",preMacroPath);
        postMacroPath = Prefs.get ("watcher.postMacroPath",postMacroPath);
        runPreMacro = Prefs.get ("watcher.runPreMacro",runPreMacro);
        runPostMacro = Prefs.get ("watcher.runPostMacro",runPostMacro);
        onlyNew = Prefs.get ("watcher.onlyNew",onlyNew);
        useBeep = Prefs.get ("watcher.useBeep",useBeep);
        showMasters = Prefs.get ("watcher.showMasters",showMasters);
        showScience = Prefs.get ("watcher.showScience",showScience);
        useGradientRemoval = Prefs.get ("watcher.useGradientRemoval",useGradientRemoval);
        useCosmicRemoval = Prefs.get ("watcher.useCosmicRemoval",useCosmicRemoval);
        showLog = Prefs.get ("watcher.showLog",showLog);
        pollingInterval = (int) Prefs.get ("watcher.pollingInterval", pollingInterval);
        dialogFrameLocationX  = (int) Prefs.get ("watcher.dialogFrameLocationX",dialogFrameLocationX);
        dialogFrameLocationY  = (int) Prefs.get ("watcher.dialogFrameLocationY",dialogFrameLocationY);
        sortNumerically = Prefs.get ("watcher.sortNumerically",sortNumerically);
        updateExclude = Prefs.get ("watcher.updateExclude",updateExclude);
        preMacro1AutoLevel = Prefs.get("watcher.preMacro1AutoLevel",preMacro1AutoLevel);
        postMacro1AutoLevel = Prefs.get("watcher.postMacro1AutoLevel",postMacro1AutoLevel);
        postMacro2AutoLevel = Prefs.get("watcher.postMacro2AutoLevel",postMacro2AutoLevel);
        minFileNumber = (long) Prefs.get ("watcher.minFileNumber",minFileNumber);
        maxFileNumber = (long) Prefs.get ("watcher.maxFileNumber",maxFileNumber);

        useBias = Prefs.get ("watcher.useBias",useBias);
        useDark = Prefs.get ("watcher.useDark",useDark);
        useFlat = Prefs.get ("watcher.useFlat",useFlat);
        useNLC = Prefs.get ("watcher.useNLC",useNLC);
        scaleExpTime = Prefs.get ("watcher.scaleExpTime",scaleExpTime);
        createBias = Prefs.get ("watcher.createBias",createBias);
        createDark = Prefs.get ("watcher.createDark",createDark);
        createFlat = Prefs.get ("watcher.createFlat",createFlat);
        saveProcessedData = Prefs.get ("watcher.saveProcessedData",saveProcessedData);
        biasMedian = Prefs.get ("watcher.biasMedian",biasMedian);
        darkMedian = Prefs.get ("watcher.darkMedian",darkMedian);
        flatMedian = Prefs.get ("watcher.flatMedian",flatMedian);
        saveFloatingPoint = Prefs.get ("watcher.saveFloatingPoint",saveFloatingPoint);
        biasDir = Prefs.get ("watcher.biasDir",biasDir);
        darkDir = Prefs.get ("watcher.darkDir",darkDir);
        flatDir = Prefs.get ("watcher.flatDir",flatDir);
        saveDir = Prefs.get ("watcher.saveDir",saveDir);
        biasBase = Prefs.get ("watcher.biasBase",biasBase);
        darkBase = Prefs.get ("watcher.darkBase",darkBase);
        flatBase = Prefs.get ("watcher.flatBase",flatBase);
        numPrefix = Prefs.get ("watcher.numPrefix",numPrefix);
        saveSuffix = Prefs.get ("watcher.saveSuffix",saveSuffix);
        saveFormat = Prefs.get ("watcher.saveFormat",saveFormat);
        coeffA = Prefs.get ("watcher.coeffA",coeffA);
        coeffB = Prefs.get ("watcher.coeffB",coeffB);
        coeffC = Prefs.get ("watcher.coeffC",coeffC);
        biasMaster = Prefs.get ("watcher.biasMaster",biasMaster);
        darkMaster = Prefs.get ("watcher.darkMaster",darkMaster);
        flatMaster = Prefs.get ("watcher.flatMaster",flatMaster);
        ra = Prefs.get ("watcher.ra",ra);
        dec = Prefs.get ("watcher.dec",dec);
        lat = Prefs.get ("watcher.lat",lat);
        lon = Prefs.get ("watcher.lon",lon);
        calcAirmass = Prefs.get ("watcher.calcAirmass",calcAirmass);
        saveAirmass = Prefs.get ("watcher.saveAirmass",saveAirmass);

        }

        void savePrefs() {

        Prefs.set ("watcher.path",path);
        Prefs.set ("watcher.filenamePattern",filenamePattern);
        Prefs.set ("watcher.preMacroPath",preMacroPath);
        Prefs.set ("watcher.postMacroPath",postMacroPath);
        Prefs.set ("watcher.runPreMacro",runPreMacro);
        Prefs.set ("watcher.runPostMacro",runPostMacro);
        Prefs.set ("watcher.onlyNew",onlyNew);
        Prefs.set ("watcher.pollingInterval",pollingInterval);
        Prefs.set ("watcher.useBeep",useBeep);
        Prefs.set ("watcher.showMasters",showMasters);
        Prefs.set ("watcher.showScience",showScience);
        Prefs.set ("watcher.useGradientRemoval",useGradientRemoval);
        Prefs.set ("watcher.useCosmicRemoval",useCosmicRemoval);
        Prefs.set ("watcher.showLog",showLog);
        Prefs.set ("watcher.sortNumerically",sortNumerically);
        Prefs.set ("watcher.updateExclude",updateExclude);
        dialogFrameLocationX = dialogFrame.getLocation().x;
        dialogFrameLocationY = dialogFrame.getLocation().y;
        Prefs.set ("watcher.dialogFrameLocationX",dialogFrameLocationX);
        Prefs.set ("watcher.dialogFrameLocationY",dialogFrameLocationY);
        Prefs.set ("watcher.preMacro1AutoLevel",preMacro1AutoLevel);
        Prefs.set ("watcher.postMacro1AutoLevel",postMacro1AutoLevel);
        Prefs.set ("watcher.postMacro2AutoLevel",postMacro2AutoLevel);
        Prefs.set ("watcher.minFileNumber",minFileNumber);
        Prefs.set ("watcher.maxFileNumber",maxFileNumber);

        Prefs.set ("watcher.useBias",useBias);
        Prefs.set ("watcher.useDark",useDark);
        Prefs.set ("watcher.useFlat",useFlat);
        Prefs.set ("watcher.useNLC",useNLC);
        Prefs.set ("watcher.scaleExpTime",scaleExpTime);
        Prefs.set ("watcher.createBias",createBias);
        Prefs.set ("watcher.createDark",createDark);
        Prefs.set ("watcher.createFlat",createFlat);
        Prefs.set ("watcher.saveProcessedData",saveProcessedData);
        Prefs.set ("watcher.biasMedian",biasMedian);
        Prefs.set ("watcher.darkMedian",darkMedian);
        Prefs.set ("watcher.flatMedian",flatMedian);
        Prefs.set ("watcher.saveFloatingPoint",saveFloatingPoint);
        Prefs.set ("watcher.biasDir",biasDir);
        Prefs.set ("watcher.darkDir",darkDir);
        Prefs.set ("watcher.flatDir",flatDir);
        Prefs.set ("watcher.saveDir",saveDir);
        Prefs.set ("watcher.biasBase",biasBase);
        Prefs.set ("watcher.darkBase",darkBase);
        Prefs.set ("watcher.flatBase",flatBase);
        Prefs.set ("watcher.numPrefix",numPrefix);
        Prefs.set ("watcher.saveSuffix",saveSuffix);
        Prefs.set ("watcher.saveFormat",saveFormat);
        Prefs.set ("watcher.coeffA",coeffA);
        Prefs.set ("watcher.coeffB",coeffB);
        Prefs.set ("watcher.coeffC",coeffC);
        Prefs.set ("watcher.biasMaster",biasMaster);
        Prefs.set ("watcher.darkMaster",darkMaster);
        Prefs.set ("watcher.flatMaster",flatMaster);
        Prefs.set ("watcher.ra",ra);
        Prefs.set ("watcher.dec",dec);
        Prefs.set ("watcher.lat",lat);
        Prefs.set ("watcher.lon",lon);
        Prefs.set ("watcher.calcAirmass",calcAirmass);
        Prefs.set ("watcher.saveAirmass",saveAirmass);
        }


	/**
	 * Starts timer which polls the target directory, looking for new images.
	 */
	protected void startTimer ()
		{
		try	{
			task = new TimerTask ()
				{
				public void run ()
					{
                    updateStatusLabel();
                    countValidFiles();
                    if (restart)
                        {
                        if (!validateSelections())
                            {
                            reset();
                            return;
                            }
                        if (createBias && !createMasterBias())
                            {
                            reset();
                            return;
                            }
                        if (createDark && !createMasterDark())
                            {
                            reset();
                            return;
                            }
                        if (createFlat && !createMasterFlat())
                            {
                            reset();
                            return;
                            }
                        restart = false;
                        }
                    if (loadNewCals && validFiles > 0)
                        {
                        updateStatusLabel();
                        countValidFiles();
                        if (!validateSelections())
                            {
                            reset();
                            return;
                            }
                        if (useBias && !loadMasterBiasFile())
                            {
                            reset();
                            return;
                            }
                        if (useDark && !loadMasterDarkFile())
                            {
                            reset();
                            return;
                            }
                        if (useFlat && !loadMasterFlatFile())
                            {
                            reset();
                            return;
                            }
                        loadNewCals = false;
                        }


					if (lockable())
						{
                        updateStatusLabel();
                        countValidFiles();
						// GET CURRENT LIST OF FILES IN DIRECTORY
                        File[] files = scienceDir.listFiles ();

                        String[] filenames = new String[files.length];
                        for (int i=0; i < files.length; i++)
                                {
                                filenames[i] = files[i].getName();
                                }

                        if (sortNumerically) filenames=sortFileList(filenames);

						for (int i=0; i < files.length; i++)
							{
							s = filenames[i];

							// NEW IMAGE?

							if ( ! images.containsKey(s))
								{
								images.put (s,null);

								// IF THE FILE LOOKS LIKE A GOOD NEW IMAGE
                                reloadFilenames = false;
								if (matchWildCard(filenamePattern, s) && stringLongVal(s) >= minFileNumber && stringLongVal(s) <= maxFileNumber)
									{
									if (useBeep) IJ.beep();
									if (onlyNew && firstRun)
										ignoredImages++;
									else
										{
										foundImages++;

                                    /* IF POLLING:
                                     * CHECK THAT FILESIZE IS CONSTANT TO VERIFY THAT FILE WRITING HAS FINISHED.
                                     * IF FILESIZE IS NOT CONSTANT, THE TIMER IS EXITED
                                     * AND A NEW TIMER IS SET UP TO TRY AGAIN */
                                    if (pollingInterval > 0)
                                        {
                                        File file = new File(path+slash+s);

                                        flen = file.length();
                                        IJ.wait(500);
                                        newflen = file.length();
                                        if (newflen != flen)
                                            {
                                            reloadFilenames = true;
                                            }
                                        }
                                    if (reloadFilenames)
                                            {
                                            images.remove(s);
                                            foundImages--;
                                            }
                                    else
                                            {
                                            if (loadNewCals)
                                                {
                                                if (!validateSelections())
                                                    {
                                                    stop();
                                                    return;
                                                    }
                                                if (useBias && !loadMasterBiasFile())
                                                    {
                                                    stop();
                                                    return;
                                                    }
                                                if (useDark && !loadMasterDarkFile())
                                                    {
                                                    stop();
                                                    return;
                                                    }
                                                if (useFlat && !loadMasterFlatFile())
                                                    {
                                                    stop();
                                                    return;
                                                    }
                                                loadNewCals = false;
                                                }
                                            sciencePath = path+s;
                                            if (showLog) IJ.log("Loading science file \""+sciencePath+"\" ("+foundImages+" of "+validFiles+")");
                                            scienceImp = IJ.openImage(sciencePath);
                                            if (scienceImp == null)
                                                    IJ.showMessage ("Unable to open image "+path+s);
                                            else
                                                    {
                                                    ImageProcessor scienceIp = scienceImp.getProcessor();

                                                    if (useBias || useDark || useFlat)
                                                        {
                                                        if (!processData())
                                                            {
                                                            stop();
                                                            return;
                                                            }
                                                        scienceIp = scienceImp.getProcessor();
                                                        }
                                                    else
                                                        {
                                                        savePath = path + saveDir.trim();
                                                        if (!savePath.endsWith(slash)) savePath += slash;
                                                        saveDirPath = savePath;
                                                        saveFileName = "Processed_"+s;

                                                        if (saveProcessedData && (saveSuffix.trim().length()!=0 || saveDir.trim().length() != 0))
                                                            {
                                                            int dotIndex = s.lastIndexOf(".");
                                                            if (dotIndex != -1) //a filetype exists
                                                                {
                                                                String filetype = s.substring(dotIndex);
                                                                saveFileName = s.replace(filetype, ""+saveSuffix.trim()+filetype);
                                                                }
                                                            else
                                                                saveFileName = s.concat(""+saveSuffix.trim()+".fits");

                                                            }
                                                        savePath += slash + saveFileName;
                                                        s = saveFileName;
                                                        scienceImp.setProcessor(s, scienceIp);
                                                        }

                                                    openImage = WindowManager.getImage(lastImageName);

                                                    if (openImage != null)
                                                        {
                                                        
                                                        openFrame = WindowManager.getFrame(lastImageName);
                                                        usingNewWindow = false;
                                                        imf = scienceImp.getFileInfo();
                                                        openImage.setFileInfo(imf);
                                                        
                                                        //CLEAR PROPERTIES FROM OPENIMAGE
                                                        props = openImage.getProperties();
                                                        if (props != null)
                                                            {
                                                            enProps = props.propertyNames();
                                                            key = "";
                                                            while (enProps.hasMoreElements())
                                                                    {
                                                                    key = (String) enProps.nextElement();
                                                                    openImage.setProperty(key, null);
                                                                    }
                                                            }
                                                        // COPY NEW PROPERTIES TO OPEN WINDOW IMAGEPLUS
                                                        props = scienceImp.getProperties();
                                                        if (props != null)
                                                            {
                                                            enProps = props.propertyNames();
                                                            key = "";
                                                            while (enProps.hasMoreElements())
                                                                    {
                                                                    key = (String) enProps.nextElement();
                                                                    openImage.setProperty(key, props.getProperty(key));
                                                                    }
                                                            }
                                                        if (openImage.getType() == ImagePlus.COLOR_RGB)
                                                            {
                                                            openImage.setDisplayRange(0, 255);
                                                            scienceIp.snapshot();
                                                            }
                                                        if (!lastImageName.equals(s))
                                                            {
                                                            testImp = WindowManager.getImage(s);
                                                            if (testImp != null) testImp.close();
                                                            }
                                                        openImage.setProcessor(s, scienceIp);
                                                        }
                                                    else
                                                        {
                                                        if (showScience)
                                                            {
                                                            usingNewWindow = true;
                                                            testImp = WindowManager.getImage(s);
                                                            if (testImp != null) testImp.close();
                                                            scienceImp.show();
                                                            if (scienceImp.getType() == ImagePlus.COLOR_RGB)
                                                                {
                                                                scienceImp.setDisplayRange(0, 255);
                                                                scienceIp.snapshot();
                                                                }
                                                            openImage = WindowManager.getImage(s);
                                                            openFrame = WindowManager.getFrame(s);
                                                            }
                                                        }

                                                    if (showScience)
                                                        {
                                                        imageWindowClass = openFrame.getClass();
                                                        if (imageWindowClass.getName().contains("AstroStackWindow"))
                                                            {
                                                            astroj.AstroStackWindow csw = (astroj.AstroStackWindow) openFrame;
                                                            csw.setAstroProcessor(true);
                                                            if (preMacro1AutoLevel) csw.setAutoLevels(s);
                                                            }

                                                        if (runPreMacro)
                                                            {
                                                            IJ.showStatus ("Running pre-macro on "+s+"...");
                                                            runner.runMacroFile (preMacroPath, s);
                                                            openImage = WindowManager.getImage(s);
                                                            openFrame = WindowManager.getFrame(s);
                                                            scienceIp = openImage.getProcessor();
                                                            }
                                                        }
                                                    if (useCosmicRemoval)
                                                        {
                                                        IJ.run(scienceImp,"Remove Outliers...","radius=2 threshold=50 which=Bright stack");
                                                        scienceIp = scienceImp.getProcessor();
                                                        scienceHeader = FitsJ.getHeader(scienceImp);
                                                        scienceHeader = FitsJ.addHistory("Hot pixels and cosmic rays removed with radius=2 and threshold=50", scienceHeader);
                                                        if (showLog) IJ.log("    Hot pixels and cosmic rays removed with radius=2 and threshold=50");
                                                        FitsJ.putHeader(scienceImp, scienceHeader);
                                                        if (showScience) FitsJ.putHeader(openImage, scienceHeader);
                                                        }
                                                    if (showScience)
                                                        {
                                                        imageWindowClass = openFrame.getClass();
                                                        if (imageWindowClass.getName().contains("AstroStackWindow"))
                                                            {
                                                            astroj.AstroStackWindow csw = (astroj.AstroStackWindow) openFrame;
                                                            csw.setAstroProcessor(true);
                                                            if (postMacro1AutoLevel) csw.setAutoLevels(s);
                                                            }

                                                        if (runPostMacro)
                                                            {
                                                            IJ.showStatus ("Running post-macro on "+s+"...");
                                                            runner.runMacroFile (postMacroPath, s);
                                                            openImage = WindowManager.getImage(s);
                                                            openFrame = WindowManager.getFrame(s);
                                                            scienceIp = openImage.getProcessor();
                                                            }
                                                        imageWindowClass = openFrame.getClass();
                                                        if (imageWindowClass.getName().contains("AstroStackWindow"))
                                                            {
                                                            astroj.AstroStackWindow csw = (astroj.AstroStackWindow) openFrame;
                                                            csw.setAstroProcessor(true);
                                                            if (postMacro2AutoLevel) csw.setAutoLevels(s);
                                                            }
                                                        }
                                                    if (saveProcessedData)
                                                        {
                                                        if (!saveProcessedFile(scienceImp, saveDirPath, savePath, "processed science", saveFormat))
                                                            {
                                                            stop();
                                                            return;
                                                            }
                                                        }
                                                    }
                                                lastImageName = s;
                                                }
                                            }
                                    }
                                else
                                    {
                                    ignoredImages++;
                                    }
                            }
                        updateStatusLabel();
                        if (requestStop) i = files.length -1;
                        }
                    firstRun = false;
                    unlock();
                }
            if (pollingInterval == 0)
                {
                reset();
                return;
                }
            }
        };
    if (!requestStop)
        {
        timer = new Timer();
        if (pollingInterval != 0) timer.schedule (task,0,pollingInterval*1000);
        else timer.schedule (task,0,5000);
                                    }
    }
catch (Exception e)
    {
    IJ.showMessage ("Error starting timer : "+e.getMessage());
    }
}


ImageProcessor normalizeImage(ImageProcessor ip, String filename)
    {

    double total = 0;
    double mean = 0;
    int height = ip.getHeight();
    int width = ip.getWidth();
    for (int y=0; y<height; y++)
        {
        for (int x=0; x<width; x++)
            {
            total += (double)(ip.getPixelValue(x, y));
            }
        }
    mean = total / ((double)width*(double)height);
    if (showLog) IJ.log("Normalizing "+filename+" with mean = "+uptoFourPlaces.format(mean));
    if (mean == 0)
        {
        return null;
        }
    for (int y=0; y<height; y++)
        {
        for (int x=0; x<width; x++)
            {
            ip.putPixelValue(x, y, (float)((double)ip.getPixelValue(x, y)/mean));
            }
        }    
    return ip;
    }

ImageProcessor removeGradient(ImageProcessor ip) { //Removes the average gradient from an image
    //
    // Adapted from c-code written by John Kielkopf (University of Louisville)
    // as part of the alsvid image processing package - specifically imlevel.c
    //
    double sy2 = 0.0, sfx = 0.0, sxy = 0.0, sfy = 0.0, sx2 = 0.0;
    double sx  = 0.0, sy  = 0.0, sf  = 0.0;
    double a, b, c, d, pixel;
    int x, y, yfits;
    double totpix = 1.0;
    int height = ip.getHeight();
    int width = ip.getWidth();
    yfits = height;
    for (y = 0; y < height; y++)    // Process one row
        {
        yfits--;
        for (x = 0; x < width; x++)   // Process each pixel in a row
            {
            pixel = ip.getPixelValue(x, yfits);
            sy2 += y*y;         // Evaluate the parameters of the least squares image gradient
            sfx += x*pixel;
            sxy += x*y;
            sfy += y*pixel;
            sx2 += x*x;
            sx += x;
            sy += y;
            sf += pixel;
            }
        }

    totpix = width*height;
    sy2 /= totpix;
    sfx /= totpix;
    sxy /= totpix;
    sfy /= totpix;
    sx2 /= totpix;
    sx  /= totpix;
    sy  /= totpix;
    sf  /= totpix;

    // Compute the gradient

    d = (sxy -sx*sy)*(sxy -sx*sy) - (sy2 -sy*sy)*(sx2 - sx*sx);
    b = (sxy - sx*sy)*(sfy - sf*sy) - (sy2 -sy*sy)*(sfx -sf*sx);
    b /= d;
    c = (sxy - sx*sy)*(sfx - sf*sx) - (sx2 -sx*sx)*(sfy -sf*sy);
    c /= d;
    a = sf - b*sx -c*sy;

//    IJ.log("tot="+totpix);
//    IJ.log("sy2="+sy2);
//    IJ.log("sfx="+sfx);
//    IJ.log("sxy="+sxy);
//    IJ.log("sfy="+sfy);
//    IJ.log("sx2="+sx2);
//    IJ.log("sx="+sx);
//    IJ.log("sy="+sy);
//    IJ.log("sf="+sf);
    if (showLog) IJ.log("    with coefficients a="+uptoFourPlaces.format(a)+
            ", b="+uptoFourPlaces.format(b)+", c="+uptoFourPlaces.format(c)+
            ", and mean="+uptoFourPlaces.format(sf));


    yfits = height;
    for (y = 0; y < height; y++)     // Loop over rows
        {
        yfits--;
        for(x = 0; x < width; x++)    //Process each pixel in a row
            {
            ip.putPixelValue(x, yfits, (ip.getPixelValue(x, yfits)  - (a + b*x + c*y - sf)));   // Remove the gradient
            }
        }
    return ip;
    }

	/** Sorts the file names into numeric order. EXTRACTED FROM ij.plugin.FolderOpener*/
	String[] sortFileList(String[] list) {
		int listLength = list.length;
		int maxDigits = 15;
		String[] list2 = null;
		char ch;
		for (int i=0; i<listLength; i++) {
			int len = list[i].length();
			String num = "";
			for (int j=0; j<len; j++) {
				ch = list[i].charAt(j);
				if (ch>=48&&ch<=57) num += ch;
			}
			if (list2==null) list2 = new String[listLength];
			if (num.length()==0) num = "aaaaaa";
			num = "000000000000000" + num; // prepend maxDigits leading zeroes
			num = num.substring(num.length()-maxDigits);
			list2[i] = num + list[i];
		}
		if (list2!=null) {
			ij.util.StringSorter.sort(list2);
			for (int i=0; i<listLength; i++)
				list2[i] = list2[i].substring(maxDigits);
			return list2;
		} else {
			ij.util.StringSorter.sort(list);
			return list;
		}
	}

    //extract the integer number within a string
    long stringLongVal(String s){
        long svalue = 0;
        String subs = s.substring(s.lastIndexOf(numPrefix)+numPrefix.length());
        int len = subs.length();
		char ch;
		String numString = "";
        for (int j=0; j<len; j++) {
            ch = subs.charAt(j);
            if (ch>=48&&ch<=57) numString += ch;
        }
		if (numString.length()!=0) svalue = Long.parseLong(numString);
        return svalue;
    }

	/**
	 * Checks to see if variables can be modified.
	 */
	synchronized protected boolean lockable ()
		{
		if (blocked == true)
			return false;
		else
			{
			blocked = true;
			return true;
			}
		}

	/**
	 * Releases control of variables.
	 */
	synchronized protected void unlock ()
		{
		blocked = false;
		}

    boolean matchWildCard(String pattern, String text)
        {
        String regex = pattern.replace("?", "[\\w\\-]").replace("*", ".*").replace("-", "\\-");
        return text.matches(regex);
        }

    public class Sexagesimal {
        private final int deg;
        private final int min;
        private final double sec;

        public Sexagesimal(int degrees, int minutes, double seconds) {
            deg = degrees;
            min = minutes;
            sec = seconds;
            }
        }


	}
